"""

analyze_classification_results.py

Given a results file in MD format, and a ground truth file in COCO Camera Traps format,
both containing classification results, perform various analyses, including:

* Precision/recall analysis
* Confusion matrix with links to visualization pages

Only analyzes image-level correctness, i.e., box locations are ignored in both
the predictions and the ground truth.

"""

#%% Imports and constants

import os
import json
import random
import argparse
import sys

import numpy as np

from collections import defaultdict
from functools import partial
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm

from megadetector.utils.path_utils import flatten_path
from megadetector.utils.write_html_image_list import write_html_image_list
from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file
from megadetector.detection.run_detector import get_typical_confidence_threshold_from_results
from megadetector.visualization import visualization_utils as vis_utils

# See "detection_category_mapping" below
default_detection_category_mapping = {}
default_detection_category_mapping['person'] = 'human'
default_detection_category_mapping['vehicle'] = 'vehicle'

# Category names treated as "no detections at all" (lowest priority).
# Any other category bumps these from a sequence's category set.
null_category_names = ['blank', 'empty']

# Category names treated as "detection with no classification" (bumps null
# categories, but is itself bumped by any non-null, non-unknown category).
unknown_category_names = ['unknown']

# Category names representing a generic "animal" detection.  Bumped by any
# animal-specific category (i.e. any category not in non_animal_category_names
# and not in null/unknown categories).
generic_animal_category_names = ['animal']

# Category names that are NOT animals.  Used to determine whether a category
# is an animal-specific prediction (which would bump generic_animal_category_names).
non_animal_category_names = ['person', 'human', 'vehicle']


#%% Support classes

class ClassificationAnalysisOptions:
    """
    Options used to parameterize analyze_classification_results().
    """

    def __init__(self):

        ### Required inputs

        #: MD-formatted results file to analyze
        self.results_file = None

        #: Ground truth file in COCO Camera Traps format
        self.gt_file = None

        ### Optional inputs

        #: Ignore all detections below this confidence threshold
        #:
        #: If this is None, a confidence threshold is selected based on the detector
        #: version.
        self.detection_threshold = None

        #: Folder where images live; filenames in [results_file] and [gt_file] should
        #: be relative to this path.  Only required if html_output_dir is not None.
        self.image_base_dir = None

        #: Folder to which we should write html output page
        self.html_output_dir = None

        #: Approximate maximum number of total images to render.  May be exceeded slightly if
        #: required to make sure that at least one image is rendered per non-empty cell in the
        #: confusion matrix.  Only relevant if html_output_dir is not None.
        self.max_total_images = 8000

        #: Try to sample this many images to render per confusion matrix cell.  Only relevant
        #: if html_output_dir is not None.  Total number is still capped by max_total_images.
        self.max_images_per_cell = 50

        #: Random seed to be used if image sampling is necessary
        self.random_seed = 0

        #: Confidence threshold to apply to classification (not detection) results
        self.classification_confidence_threshold = 0.6

        #: A dict mapping detection category names to classification category names, for
        #: categories we want to handle specially.  Any detection in a matching category with
        #: an above-threshold confidence value is treated as if it had a classification
        #: with the corresponding (mapped) classification category, with a confidence of 1.0,
        #: whether or not that category exists in the ground truth.
        #:
        #: For example, by default a detection with category "person" with confidence 0.4
        #: should be treated as a classification of category "human" with confidence 1.0.
        #:
        #: Defaults to detection_category_mapping.
        self.detection_category_mapping = None

        #: When this is True (default), we trust detection categories regardless of
        #: classifications.  I.e., a detection category of "person" with a classification
        #: of "elk" will be treated as a classification of "person".
        #:
        #: When this is False, we trust classifications, if present.  This is generally
        #: used when I'm simulating Wildlife Insights ensemble behavior, where "classifications"
        #: at this point really represent the output of the entire WI ensemble.
        self.apply_detection_category_mapping_when_classifications_are_present = True

        #: If True, the entire analysis will be performed at the *sequence* level, rather
        #: than the image level.
        self.sequence_level_analysis = False

        #: Number of workers to use when rendering images
        self.rendering_workers = 10

        #: Should we use threads ("threads") or processes ("processes") for rendering?
        #:
        #: Only relevant if rendering_workers is > 1.
        self.rendering_pool_type = 'threads'

        #: Should we over-write images that already exist?
        self.overwrite = True

        # Should we show the "overall metrics" table?
        self.show_overall_metrics = True

        #: Width of rendered output images (-1 to preserve original size)
        self.output_image_width = 1000

        #: Number of top misprediction categories to show in the per-category
        #: statistics table (for both "mispredicted as this" and "this was
        #: mispredicted as" columns).
        self.n_mispredictions_for_table = 5

        #: List of category names to completely exclude from the analysis,
        #: whether they appear in ground truth or predictions.  Checked after
        #: detection category remapping (e.g. person -> human).
        self.categories_to_ignore = None

        #: If True, collapse each image's (or sequence's) predictions to a
        #: single category.  The winner is chosen by the *number* of above-
        #: threshold classifications for that category, with max classification
        #: confidence as a tie-breaker.  For sequence-level analyses, collapsing
        #: happens at the sequence level (counts are summed across images in
        #: the sequence), not per-image.
        self.single_prediction_per_image = False

        #: If True, collapse each image's (or sequence's) ground truth to a
        #: single category.  For image-level analyses, ties are broken
        #: alphabetically.  For sequence-level analyses, ties are broken by
        #: count (how many images in the sequence have that category), then
        #: alphabetically.
        self.single_label_per_image = False

        #: Maximum number of images to include in any HTML page, generally only
        #: relevant when calling render_misprediction_pages(...).
        self.max_images_per_html_file = 1000

        #: Before processing, optionally map a subset of predicted classification categories
        #: to alternative names.  Typically used to reconcile category names across predictions/GT.
        #:
        #: If not None, should be a str --> str dict.
        self.predicted_category_name_mappings = None

        #: Before processing, optionally map a subset of GT classification categories
        #: to alternative names.  Typically used to reconcile category names across predictions/GT.
        #:
        #: If not None, should be a str --> str dict.
        self.gt_category_name_mappings = None

    # ...def __init__(...)

# ...class ClassificationAnalysisOptions


class AnalysisResults:
    """
    Results returned by analyze_classification_results().
    """

    def __init__(self):

        #: Dictionary mapping category names to dicts, where each item has
        #: at least the keys "precision", "recall", "f1", "n_ground_truth", "n_predicted"
        self.per_category_results = None

        self.macro_f1 = None
        self.micro_f1 = None
        self.micro_precision = None
        self.micro_recall = None
        self.accuracy = None

        #: The confusion matrix as a 2D numpy array
        self.confusion_matrix = None

        #: Ordered list of category names corresponding to matrix rows/columns
        self.active_categories = None

        #: Path to the output HTML file (if generated)
        self.html_output_file = None

    # ...def __init__(...)

# ...class AnalysisResults


#%% Support functions

def _collapse_sequence_categories(categories):
    """
    Collapse a set of categories by removing less-specific categories when
    more-specific ones are present, according to the priority hierarchy:

        null (blank/empty) < unknown < generic animal / non-animal specifics

    Also, generic animal categories (e.g. "animal") are bumped by any
    animal-specific category.

    Args:
        categories (set): set of lowercase category name strings

    Returns:
        set: the collapsed category set
    """

    if len(categories) <= 1:
        return categories

    cats = set(categories)

    null_set = set(null_category_names)
    unknown_set = set(unknown_category_names)
    generic_animal_set = set(generic_animal_category_names)
    non_animal_set = set(non_animal_category_names)

    all_special = null_set | unknown_set | generic_animal_set

    # Check what tiers are present
    has_non_null_non_unknown = any(c not in null_set and c not in unknown_set for c in cats)
    has_specific_animal = any(
        c not in all_special and c not in non_animal_set for c in cats)

    # Null categories are bumped by anything else
    if has_non_null_non_unknown or (cats & unknown_set):
        cats -= null_set

    # Unknown categories are bumped by anything that isn't null or unknown
    if has_non_null_non_unknown:
        cats -= unknown_set

    # Generic animal is bumped by any animal-specific category
    if has_specific_animal:
        cats -= generic_animal_set

    return cats

# ...def _collapse_sequence_categories(...)


def _get_image_predicted_categories(im,
                                    detection_category_id_to_name,
                                    classification_category_id_to_name,
                                    detection_category_mapping,
                                    options):
    """
    For a single MD results image entry, returns a tuple of:

    - a dict mapping predicted category names (lowercase) to their max confidence
    - a dict mapping predicted category names (lowercase) to the number of
      above-threshold classifications for that category
    """

    predicted_categories = {}
    predicted_counts = defaultdict(int)

    has_above_threshold_detection = False

    if im.get('detections') is None:
        return {'empty': 1.0}, {'empty': 1}

    for det in im['detections']:

        if det['conf'] < options.detection_threshold:
            continue

        has_above_threshold_detection = True

        det_category_name = detection_category_id_to_name.get(det['category'], None)
        if det_category_name is None:
            continue

        det_category_name_lower = det_category_name.lower()

        # Do we need to look at classifications?  This will be False if we infer the category
        # from detections alone.
        review_classifications = True

        # Determine whether the detection category should be used as the prediction
        # for this object
        if (det_category_name_lower in detection_category_mapping):

            # If we have classifications and we're supposed to use them in
            # this scenario...
            if ('classifications' in det) and \
                (len(det['classifications']) > 0) and \
                (not options.apply_detection_category_mapping_when_classifications_are_present):
                review_classifications = True
            else:
                mapped_name = detection_category_mapping[det_category_name_lower].lower()
                if mapped_name not in predicted_categories or predicted_categories[mapped_name] < 1.0:
                    predicted_categories[mapped_name] = 1.0
                predicted_counts[mapped_name] += 1
                review_classifications = False

        if review_classifications:

            # Look at classifications
            classifications = det.get('classifications', None)
            has_above_threshold_classification = False

            if classifications is not None:
                for cls in classifications:
                    cls_id = str(cls[0])
                    cls_conf = cls[1]
                    if cls_conf >= options.classification_confidence_threshold:
                        cls_name = classification_category_id_to_name.get(cls_id, None)
                        if cls_name is not None:
                            cls_name_lower = cls_name.lower()
                            has_above_threshold_classification = True
                            if cls_name_lower not in predicted_categories or \
                                    predicted_categories[cls_name_lower] < cls_conf:
                                predicted_categories[cls_name_lower] = cls_conf
                            predicted_counts[cls_name_lower] += 1

            if not has_above_threshold_classification:
                if 'unknown' not in predicted_categories:
                    predicted_categories['unknown'] = 1.0
                predicted_counts['unknown'] += 1

        # ...if we're supposed to look at classifications for this object

    # ...for each detection

    if not has_above_threshold_detection:
        return {'empty': 1.0}, {'empty': 1}

    return predicted_categories, dict(predicted_counts)

# ...def _get_image_predicted_categories(...)


def _render_single_image(im, render_constants):
    """
    Renders a single image with bounding boxes to the output folder.

    Args:
        im (dict): image entry from MD results, with an added '_output_path' key
        render_constants (dict): rendering parameters

    Returns:
        dict: result with at least the keys 'success', 'output_path', 'file', and
        'error'
    """

    result = {
        'success': False,
        'output_path': None,
        'error': None,
        'file': im['file']
    }

    # Validate inputs
    if im['detections'] is None:

        failure_string = ''
        if 'failure' in im:
            failure_string = im['failure']
        print('Warning: skipping rendering for {}: failure string [{}]'.format(
            im['file'],failure_string))
        result['error'] = 'inference failure'
        return result

    # Expand input dict to local variables
    image_base_dir = render_constants['image_base_dir']
    output_dir = render_constants['output_dir']
    detection_label_map = render_constants['detection_label_map']
    classification_label_map = render_constants['classification_label_map']
    detection_threshold = render_constants['detection_threshold']
    classification_confidence_threshold = render_constants['classification_confidence_threshold']
    output_image_width = render_constants['output_image_width']
    overwrite = render_constants['overwrite']

    # Build output filename
    fn_clean = flatten_path(im['file']).replace(' ', '_')
    output_path = os.path.join(output_dir, fn_clean)

    # Skip if already rendered
    if os.path.isfile(output_path) and (not overwrite):

        result['success'] = True
        result['output_path'] = output_path
        result['error'] = 'Skipped rendering, image exists'
        return result

    input_path = os.path.join(image_base_dir, im['file'])

    if not os.path.isfile(input_path):
        print('Warning: image not found: {}'.format(input_path))
        result['error'] = 'image not found'
        return result

    image = vis_utils.open_image(input_path)
    image = vis_utils.resize_image(image, output_image_width)

    vis_utils.render_detection_bounding_boxes(
        im['detections'],
        image,
        label_map=detection_label_map,
        classification_label_map=classification_label_map,
        confidence_threshold=detection_threshold,
        classification_confidence_threshold=classification_confidence_threshold)

    image.save(output_path)
    result['success'] = True
    result['output_path'] = output_path

    return result

# ...def _render_single_image(...)


#%% Core functions

def _prepare_analysis_data(options):
    """
    Load results and ground truth files, build category assignments, apply filtering
    and aggregation options, and return the prepared data needed for analysis and
    rendering.

    Args:
        options (ClassificationAnalysisOptions): options object defining filenames
            and analysis parameters.

    Returns:
        dict: prepared analysis data with keys:
            - filename_to_gt_categories
            - filename_to_pred_categories
            - filename_to_pred_counts
            - active_categories
            - category_to_index
            - results_fn_to_im
            - gt_data
            - detection_category_id_to_name
            - classification_category_id_to_name
            - detection_threshold
            - categories_to_ignore
    """

    ## Setup and defaults

    if options.detection_category_mapping is None:
        detection_category_mapping = {k.lower(): v.lower()
                                      for k, v in default_detection_category_mapping.items()}
    else:
        detection_category_mapping = {k.lower(): v.lower()
                                      for k, v in options.detection_category_mapping.items()}

    ## Load files

    print('Loading results from {}'.format(options.results_file))
    results_data = load_md_or_speciesnet_file(options.results_file)

    detection_threshold = options.detection_threshold
    if detection_threshold is None:
        detection_threshold = get_typical_confidence_threshold_from_results(results_data)
        print('Using auto-detected confidence threshold: {}'.format(detection_threshold))

    print('Loading ground truth from {}'.format(options.gt_file))
    with open(options.gt_file, 'r') as f:
        gt_data = json.load(f)

    # Build category maps
    detection_category_id_to_name = results_data.get('detection_categories', {})
    classification_category_id_to_name = results_data.get('classification_categories', {})

    gt_category_id_to_name = {}
    for c in gt_data['categories']:
        gt_category_id_to_name[c['id']] = c['name']

    # Normalize filenames (backslash -> forward slash)
    for im in results_data['images']:
        im['file'] = im['file'].replace('\\', '/')

    for im in gt_data['images']:
        im['file_name'] = im['file_name'].replace('\\', '/')

    ## Validate file compatibility

    results_fn_to_im = {}
    for im in results_data['images']:
        results_fn_to_im[im['file']] = im

    gt_fn_to_im = {}
    for im in gt_data['images']:
        gt_fn_to_im[im['file_name']] = im

    common_filenames = set(results_fn_to_im.keys()) & set(gt_fn_to_im.keys())
    results_only = set(results_fn_to_im.keys()) - set(gt_fn_to_im.keys())
    gt_only = set(gt_fn_to_im.keys()) - set(results_fn_to_im.keys())

    print('{} images in common between results and ground truth'.format(len(common_filenames)))

    if len(results_only) > 0:
        print('Warning: {} images in results but not in ground truth (GT has {}, results have {})'.format(
            len(results_only),len(gt_fn_to_im),len(results_fn_to_im)))
    if len(gt_only) > 0:
        print('Warning: {} images in ground truth but not in results (GT has {}, results have {})'.format(
            len(gt_only),len(gt_fn_to_im),len(results_fn_to_im)))

    assert len(common_filenames) > 0, \
        'No images in common between results and ground truth'

    ## Build GT category assignments

    gt_image_id_to_annotations = defaultdict(list)
    for ann in gt_data['annotations']:
        gt_image_id_to_annotations[ann['image_id']].append(ann)

    filename_to_gt_categories = {}
    n_images_without_annotations = 0

    for im in gt_data['images']:

        fn = im['file_name']
        if fn not in common_filenames:
            continue

        annotations = gt_image_id_to_annotations.get(im['id'], [])
        if len(annotations) == 0:
            n_images_without_annotations += 1
            continue

        gt_cats = set()
        for ann in annotations:
            cat_name = gt_category_id_to_name[ann['category_id']].lower()
            if options.gt_category_name_mappings is not None:
                cat_name = options.gt_category_name_mappings.get(cat_name, cat_name)
            gt_cats.add(cat_name)

        filename_to_gt_categories[fn] = gt_cats

    # ...for each filename

    if n_images_without_annotations > 0:
        print('Warning: {} images in GT with no annotations (excluded)'.format(
            n_images_without_annotations))

    print('{} images with ground truth annotations'.format(len(filename_to_gt_categories)))

    ## Build predicted category assignments

    filename_to_pred_categories = {}
    filename_to_pred_counts = {}

    for fn in filename_to_gt_categories:

        im = results_fn_to_im[fn]

        pred_cats, pred_counts = _get_image_predicted_categories(
            im,
            detection_category_id_to_name,
            classification_category_id_to_name,
            detection_category_mapping,
            options)

        if options.predicted_category_name_mappings is not None:

            mapped_cats = {}
            mapped_counts = {}
            for cat_name, conf in pred_cats.items():
                mapped_name = options.predicted_category_name_mappings.get(cat_name, cat_name)
                mapped_cats[mapped_name] = max(conf, mapped_cats.get(mapped_name, 0))
            for cat_name, count in pred_counts.items():
                mapped_name = options.predicted_category_name_mappings.get(cat_name, cat_name)
                mapped_counts[mapped_name] = count + mapped_counts.get(mapped_name, 0)
            pred_cats = mapped_cats
            pred_counts = mapped_counts

        filename_to_pred_categories[fn] = pred_cats
        filename_to_pred_counts[fn] = pred_counts

    # ...for each filename

    ## Filter ignored categories

    categories_to_ignore = set()
    if options.categories_to_ignore is not None:
        categories_to_ignore = set(c.lower() for c in options.categories_to_ignore)

    if len(categories_to_ignore) > 0:

        print('Ignoring {} categories: {}'.format(
            len(categories_to_ignore), sorted(categories_to_ignore)))

        # Remove ignored categories from GT
        fns_to_remove = []
        for fn in filename_to_gt_categories:
            filename_to_gt_categories[fn] -= categories_to_ignore
            if len(filename_to_gt_categories[fn]) == 0:
                fns_to_remove.append(fn)

        for fn in fns_to_remove:
            del filename_to_gt_categories[fn]
            del filename_to_pred_categories[fn]
            del filename_to_pred_counts[fn]

        if len(fns_to_remove) > 0:
            print('  {} images excluded because all their GT categories were ignored'.format(
                len(fns_to_remove)))

        # Remove ignored categories from predictions
        for fn in filename_to_pred_categories:
            for cat in categories_to_ignore:
                filename_to_pred_categories[fn].pop(cat, None)
                filename_to_pred_counts[fn].pop(cat, None)

        print('{} images remaining after filtering'.format(len(filename_to_gt_categories)))

    # ...if we're supposed to ignore some categories

    ## Sequence-level aggregation

    if options.sequence_level_analysis:

        # Build filename -> seq_id map from GT
        filename_to_seq_id = {}
        for im in gt_data['images']:
            fn = im['file_name']
            if fn in filename_to_gt_categories:
                assert 'seq_id' in im, \
                    'Image {} has no seq_id; sequence-level analysis requires seq_id'.format(fn)
                filename_to_seq_id[fn] = im['seq_id']

        # Group filenames by seq_id
        seq_id_to_filenames = defaultdict(list)
        for fn, seq_id in filename_to_seq_id.items():
            seq_id_to_filenames[seq_id].append(fn)

        # Aggregate GT and predictions per sequence
        seq_filename_to_gt_categories = {}
        seq_filename_to_pred_categories = {}
        seq_filename_to_pred_counts = {}

        for seq_id, filenames in seq_id_to_filenames.items():

            gt_union = set()
            pred_union = {}
            pred_count_union = defaultdict(int)
            for fn in filenames:
                gt_union |= filename_to_gt_categories[fn]
                for cat, conf in filename_to_pred_categories[fn].items():
                    if cat not in pred_union or pred_union[cat] < conf:
                        pred_union[cat] = conf
                for cat, count in filename_to_pred_counts[fn].items():
                    pred_count_union[cat] += count

            # Collapse category hierarchies for the sequence; e.g. if a sequence
            # has both "empty" and "deer" predictions, remove "empty".
            collapsed_gt = _collapse_sequence_categories(gt_union)
            gt_union = collapsed_gt

            collapsed_pred = _collapse_sequence_categories(set(pred_union.keys()))
            removed_pred_cats = set(pred_union.keys()) - collapsed_pred
            for cat in removed_pred_cats:
                del pred_union[cat]
                pred_count_union.pop(cat, None)

            seq_filename_to_gt_categories[seq_id] = gt_union
            seq_filename_to_pred_categories[seq_id] = pred_union
            seq_filename_to_pred_counts[seq_id] = dict(pred_count_union)

        # ...for each sequence

        # Replace image-level dicts with sequence-level dicts
        filename_to_gt_categories = seq_filename_to_gt_categories
        filename_to_pred_categories = seq_filename_to_pred_categories
        filename_to_pred_counts = seq_filename_to_pred_counts

        print('{} sequences for analysis'.format(len(filename_to_gt_categories)))

    # ...if we're doing a sequence-level analysis

    ## Collapse ground truth to single label per image/sequence

    if options.single_label_per_image:

        for entity_id in filename_to_gt_categories:

            gt_cats = filename_to_gt_categories[entity_id]
            if len(gt_cats) <= 1:
                continue

            if options.sequence_level_analysis:
                # Break ties by count (how many images have this category), then
                # alphabetically.
                seq_id = entity_id
                filenames_in_seq = seq_id_to_filenames[seq_id]
                cat_counts = defaultdict(int)
                for fn in filenames_in_seq:
                    # Use the original per-image GT (before sequence aggregation),
                    # but we already replaced that dict.  We can recount from
                    # gt_image_id_to_annotations.
                    gt_im = gt_fn_to_im[fn]
                    anns = gt_image_id_to_annotations.get(gt_im['id'], [])
                    for ann in anns:
                        cat_name = gt_category_id_to_name[ann['category_id']].lower()
                        if cat_name not in categories_to_ignore:
                            cat_counts[cat_name] += 1
                # Pick by count (descending), then alphabetically
                winner = max(gt_cats, key=lambda c: (cat_counts.get(c, 0), -ord(c[0])))
                # More robust: sort by (-count, name)
                winner = sorted(gt_cats, key=lambda c: (-cat_counts.get(c, 0), c))[0]
            else:
                # Image-level: break ties alphabetically
                winner = sorted(gt_cats)[0]

            filename_to_gt_categories[entity_id] = {winner}

        # ...for each image or sequence

        print('Collapsed ground truth to single label per {}'.format(
            'sequence' if options.sequence_level_analysis else 'image'))

    # ...if we're supposed to collapse ground truth to a single label per image/sequence

    ## Collapse predictions to single label per image/sequence

    if options.single_prediction_per_image:

        for entity_id in filename_to_pred_categories:

            pred_cats = filename_to_pred_categories[entity_id]
            if len(pred_cats) <= 1:
                continue

            counts = filename_to_pred_counts[entity_id]

            # Pick by count (descending), then max confidence (descending),
            # then alphabetically
            winner = sorted(
                pred_cats.keys(),
                key=lambda c: (-counts.get(c, 0), -pred_cats[c], c))[0]

            winner_conf = pred_cats[winner]
            filename_to_pred_categories[entity_id] = {winner: winner_conf}
            filename_to_pred_counts[entity_id] = {winner: counts.get(winner, 1)}

        print('Collapsed predictions to single label per {}'.format(
            'sequence' if options.sequence_level_analysis else 'image'))

        # ...for each image or sequence

    # ...if we are collapsing to a single prediction per image

    ## Determine active categories

    all_gt_categories = set()
    for cats in filename_to_gt_categories.values():
        all_gt_categories |= cats

    all_pred_categories = set()
    for cats in filename_to_pred_categories.values():
        all_pred_categories |= set(cats.keys())

    active_categories = sorted(all_gt_categories | all_pred_categories)

    print('{} active categories'.format(len(active_categories)))

    category_to_index = {cat: i for i, cat in enumerate(active_categories)}

    return {
        'filename_to_gt_categories': filename_to_gt_categories,
        'filename_to_pred_categories': filename_to_pred_categories,
        'filename_to_pred_counts': filename_to_pred_counts,
        'active_categories': active_categories,
        'category_to_index': category_to_index,
        'results_fn_to_im': results_fn_to_im,
        'gt_data': gt_data,
        'detection_category_id_to_name': detection_category_id_to_name,
        'classification_category_id_to_name': classification_category_id_to_name,
        'detection_threshold': detection_threshold,
        'categories_to_ignore': categories_to_ignore
    }

# ...def _prepare_analysis_data(...)


def analyze_classification_results(options):
    """
    Perform precision-recall analysis on classification results.

    Args:
        options (ClassificationAnalysisOptions): options object defining filenames
            and analysis parameters.

    Returns:
        AnalysisResults: results of the classification analysis
    """

    prepared = _prepare_analysis_data(options)

    filename_to_gt_categories = prepared['filename_to_gt_categories']
    filename_to_pred_categories = prepared['filename_to_pred_categories']
    active_categories = prepared['active_categories']
    category_to_index = prepared['category_to_index']
    results_fn_to_im = prepared['results_fn_to_im']
    gt_data = prepared['gt_data']
    detection_category_id_to_name = prepared['detection_category_id_to_name']
    classification_category_id_to_name = prepared['classification_category_id_to_name']
    detection_threshold = prepared['detection_threshold']
    categories_to_ignore = prepared['categories_to_ignore']

    ## Build confusion matrix

    n_categories = len(active_categories)
    confusion_matrix = np.zeros((n_categories, n_categories), dtype=int)

    # Maps (true_cat, pred_cat) -> list of filenames/seq_ids
    true_pred_to_filenames = defaultdict(list)

    for entity_id in filename_to_gt_categories:

        gt_cats = filename_to_gt_categories[entity_id]
        pred_cats = filename_to_pred_categories[entity_id]

        for true_cat in gt_cats:

            for pred_cat in pred_cats:

                # For off-diagonal entries, skip cases where both categories are
                # correctly present, i.e. the predicted category is in GT and the
                # true category was also predicted.  In this case the cross-product
                # would create a spurious "misprediction" entry.
                if true_cat != pred_cat:
                    if pred_cat in gt_cats and true_cat in pred_cats:
                        continue

                true_idx = category_to_index[true_cat]
                pred_idx = category_to_index[pred_cat]
                confusion_matrix[true_idx, pred_idx] += 1
                true_pred_to_filenames[(true_cat, pred_cat)].append(entity_id)

            # ...for each predicted category

        # ...for each true category

    # ...for each image or sequence

    ## Compute metrics

    # Count entities (images/sequences) per category, independent of the
    # confusion matrix.  N(GT) = number of entities where this category is in
    # the ground truth.  N(Pred) = number of entities where this category was
    # predicted at least once with above-threshold confidence.
    gt_entity_counts = defaultdict(int)
    pred_entity_counts = defaultdict(int)

    for entity_id in filename_to_gt_categories:
        for cat in filename_to_gt_categories[entity_id]:
            gt_entity_counts[cat] += 1
        for cat in filename_to_pred_categories[entity_id]:
            pred_entity_counts[cat] += 1

    per_category_results = {}
    total_tp = 0
    total_predicted = 0
    total_gt = 0

    for cat in active_categories:

        idx = category_to_index[cat]
        tp = confusion_matrix[idx, idx]
        fp = int(confusion_matrix[:, idx].sum()) - tp
        fn = int(confusion_matrix[idx, :].sum()) - tp

        n_gt = gt_entity_counts[cat]
        n_pred = pred_entity_counts[cat]

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = (2 * precision * recall) / (precision + recall) if ((precision + recall) > 0) else 0.0

        per_category_results[cat] = {
            'precision': precision,
            'recall': recall,
            'f1': f1,
            'n_ground_truth': int(n_gt),
            'n_predicted': int(n_pred),
            'tp': int(tp),
            'fp': int(fp),
            'fn': int(fn)
        }

        total_tp += tp
        total_predicted += (tp + fp)
        total_gt += (tp + fn)

    # ...for each category

    # Compute per-image micro precision and recall.
    #
    # When computed from the confusion matrix, micro precision always
    # equals micro recall (because total column sums == total row sums ==
    # total matrix entries).  Instead, we compute per-image set-based metrics:
    #
    #   micro_precision = sum(|GT ∩ pred|) / sum(|pred|)
    #   micro_recall    = sum(|GT ∩ pred|) / sum(|GT|)
    #
    # These can differ when images have multiple GT and/or predicted categories.

    sum_intersection = 0
    sum_gt_size = 0
    sum_pred_size = 0

    for entity_id in filename_to_gt_categories:

        gt_cats = filename_to_gt_categories[entity_id]
        pred_cats_set = set(filename_to_pred_categories[entity_id].keys())
        sum_intersection += len(gt_cats & pred_cats_set)
        sum_gt_size += len(gt_cats)
        sum_pred_size += len(pred_cats_set)

    micro_precision = sum_intersection / sum_pred_size if sum_pred_size > 0 else 0.0
    micro_recall = sum_intersection / sum_gt_size if sum_gt_size > 0 else 0.0
    micro_f1 = (2 * micro_precision * micro_recall) / (micro_precision + micro_recall) \
        if (micro_precision + micro_recall) > 0 else 0.0

    f1_values = [per_category_results[cat]['f1'] for cat in active_categories
                 if per_category_results[cat]['n_ground_truth'] > 0]
    macro_f1 = np.mean(f1_values) if len(f1_values) > 0 else 0.0

    matrix_sum = confusion_matrix.sum()
    diagonal_sum = np.trace(confusion_matrix)
    accuracy = diagonal_sum / matrix_sum if matrix_sum > 0 else 0.0

    print('\nOverall metrics:')
    print('  Accuracy: {:.4f}'.format(accuracy))
    print('  Macro F1: {:.4f}'.format(macro_f1))
    print('  Micro precision: {:.4f}'.format(micro_precision))
    print('  Micro recall: {:.4f}'.format(micro_recall))
    print('  Micro F1: {:.4f}'.format(micro_f1))

    ## Generate HTML output

    html_output_file = None

    if options.html_output_dir is not None:

        assert options.image_base_dir is not None, \
            'image_base_dir is required when html_output_dir is specified'

        os.makedirs(options.html_output_dir, exist_ok=True)
        preview_images_folder = os.path.join(options.html_output_dir, 'images')
        os.makedirs(preview_images_folder, exist_ok=True)

        # Determine which images to render

        # For sequence-level analysis, we need to map back to individual filenames
        if options.sequence_level_analysis:
            seq_id_to_filenames_map = defaultdict(list)
            for im in gt_data['images']:
                fn = im['file_name']
                if fn in results_fn_to_im and 'seq_id' in im:
                    seq_id_to_filenames_map[im['seq_id']].append(fn)

        # Collect non-empty cells and their filenames
        non_empty_cells = {}
        for (true_cat, pred_cat), entity_ids in true_pred_to_filenames.items():
            if len(entity_ids) > 0:
                non_empty_cells[(true_cat, pred_cat)] = entity_ids

        # Build misprediction data with separate FP and FN perspectives.
        #
        # FP perspective ("mispredicted as this category"): for each predicted
        # category that is NOT in the ground truth, pair with each GT category.
        # mispred_fp[pred_cat][true_cat] -> [entity_ids]
        mispred_fp = defaultdict(lambda: defaultdict(list))
        #
        # FN perspective ("mispredicted as"): for each GT category that is NOT
        # among the predictions, pair with each predicted category.
        # mispred_fn[true_cat][pred_cat] -> [entity_ids]
        mispred_fn = defaultdict(lambda: defaultdict(list))

        for entity_id in filename_to_gt_categories:
            gt_cats = filename_to_gt_categories[entity_id]
            pred_cats = set(filename_to_pred_categories[entity_id].keys())

            # FP: predicted categories not in GT
            wrong_preds = pred_cats - gt_cats
            for pred_cat in wrong_preds:
                for true_cat in gt_cats:
                    mispred_fp[pred_cat][true_cat].append(entity_id)

            # FN: GT categories not in predictions
            missed_gts = gt_cats - pred_cats
            for true_cat in missed_gts:
                for pred_cat in pred_cats:
                    mispred_fn[true_cat][pred_cat].append(entity_id)

        # Build top-N summaries and collect entity lists for per-cell pages
        mispredicted_as_this = {}
        fp_entities = {}  # (pred_cat, true_cat) -> [entity_ids]

        this_was_mispredicted_as = {}
        fn_entities = {}  # (true_cat, pred_cat) -> [entity_ids]

        n_mispred_for_mispred = options.n_mispredictions_for_table

        for cat in active_categories:

            col_entries = []
            for other_cat, entity_ids in mispred_fp.get(cat, {}).items():
                if len(entity_ids) > 0:
                    col_entries.append((other_cat, len(entity_ids)))
                    fp_entities[(cat, other_cat)] = entity_ids
            col_entries.sort(key=lambda x: -x[1])
            mispredicted_as_this[cat] = col_entries[:n_mispred_for_mispred]

            row_entries = []
            for other_cat, entity_ids in mispred_fn.get(cat, {}).items():
                if len(entity_ids) > 0:
                    row_entries.append((other_cat, len(entity_ids)))
                    fn_entities[(cat, other_cat)] = entity_ids
            row_entries.sort(key=lambda x: -x[1])
            this_was_mispredicted_as[cat] = row_entries[:n_mispred_for_mispred]

        # ...for each category

        # Determine which misprediction cells need pages (FP and FN separately)
        fp_non_empty_cells = {}
        for (pred_cat, true_cat), entity_ids in fp_entities.items():
            # Only generate pages for cells that appear in the top-N summaries
            if any(true_cat == other_cat
                   for other_cat, _ in mispredicted_as_this.get(pred_cat, [])):
                fp_non_empty_cells[('fp', pred_cat, true_cat)] = entity_ids

        fn_non_empty_cells = {}
        for (true_cat, pred_cat), entity_ids in fn_entities.items():
            if any(pred_cat == other_cat
                   for other_cat, _ in this_was_mispredicted_as.get(true_cat, [])):
                fn_non_empty_cells[('fn', true_cat, pred_cat)] = entity_ids

        # Determine how many images to sample per cell across regular, FP
        # misprediction, and FN misprediction cells.  First cap each at
        # max_images_per_cell, then scale down proportionally if the combined
        # total exceeds max_total_images.
        cell_desired_counts = {}
        for key, entity_ids in non_empty_cells.items():
            cell_desired_counts[key] = min(len(entity_ids), options.max_images_per_cell)
        for key, entity_ids in fp_non_empty_cells.items():
            cell_desired_counts[key] = min(len(entity_ids), options.max_images_per_cell)
        for key, entity_ids in fn_non_empty_cells.items():
            cell_desired_counts[key] = min(len(entity_ids), options.max_images_per_cell)

        total_desired = sum(cell_desired_counts.values())

        if total_desired > options.max_total_images and total_desired > 0:
            scale = options.max_total_images / total_desired
            for key in cell_desired_counts:
                cell_desired_counts[key] = max(1, int(cell_desired_counts[key] * scale))

        # Sample entities per cell
        rng = random.Random(options.random_seed)
        sampled_cells = {}

        for key, entity_ids in non_empty_cells.items():
            n_sample = cell_desired_counts[key]
            if len(entity_ids) <= n_sample:
                sampled_cells[key] = list(entity_ids)
            else:
                sampled_cells[key] = rng.sample(entity_ids, n_sample)

        sampled_fp_cells = {}
        for key, entity_ids in fp_non_empty_cells.items():
            n_sample = cell_desired_counts[key]
            if len(entity_ids) <= n_sample:
                sampled_fp_cells[(key[1], key[2])] = list(entity_ids)
            else:
                sampled_fp_cells[(key[1], key[2])] = rng.sample(entity_ids, n_sample)

        sampled_fn_cells = {}
        for key, entity_ids in fn_non_empty_cells.items():
            n_sample = cell_desired_counts[key]
            if len(entity_ids) <= n_sample:
                sampled_fn_cells[(key[1], key[2])] = list(entity_ids)
            else:
                sampled_fn_cells[(key[1], key[2])] = rng.sample(entity_ids, n_sample)

        # Collect all unique filenames that need rendering.
        #
        # For sequence-level analysis, render only the first image in each
        # sequence as an exemplar.
        filenames_to_render = set()

        def _collect_filenames(entity_id_lists):
            for entity_ids in entity_id_lists:
                if options.sequence_level_analysis:
                    for seq_id in entity_ids:
                        fns = seq_id_to_filenames_map.get(seq_id, [])
                        if len(fns) > 0:
                            filenames_to_render.add(fns[0])
                else:
                    filenames_to_render.update(entity_ids)

        _collect_filenames(sampled_cells.values())
        _collect_filenames(sampled_fp_cells.values())
        _collect_filenames(sampled_fn_cells.values())

        # Build image entries for rendering
        images_to_render = []
        for fn in filenames_to_render:
            if fn in results_fn_to_im:
                images_to_render.append(results_fn_to_im[fn])

        print('\nRendering {} images...'.format(len(images_to_render)))

        # Render images
        render_constants = {
            'image_base_dir': options.image_base_dir,
            'output_dir': preview_images_folder,
            'detection_label_map': detection_category_id_to_name,
            'classification_label_map': classification_category_id_to_name,
            'detection_threshold': detection_threshold,
            'classification_confidence_threshold': options.classification_confidence_threshold,
            'output_image_width': options.output_image_width,
            'overwrite': options.overwrite
        }

        if options.rendering_workers > 1 and len(images_to_render) > 1:

            if options.rendering_pool_type == 'threads':
                pool = ThreadPool(options.rendering_workers)
                worker_string = 'threads'
            else:
                pool = Pool(options.rendering_workers)
                worker_string = 'processes'

            print('Rendering with {} {}'.format(options.rendering_workers, worker_string))

            try:
                rendering_results = list(tqdm(
                    pool.imap(
                        partial(_render_single_image, render_constants=render_constants),
                        images_to_render),
                    total=len(images_to_render)))
            finally:
                pool.close()
                pool.join()
        else:

            rendering_results = []
            for im in tqdm(images_to_render):
                rendering_results.append(_render_single_image(im, render_constants))

        # ...if we are parallelizing rendering

        n_success = sum(1 for r in rendering_results if r['success'])
        print('Successfully rendered {} of {} images'.format(n_success, len(images_to_render)))

        # Generate per-cell HTML pages

        for (true_cat, pred_cat), entity_ids in sampled_cells.items():

            html_image_info_list = []

            for entity_id in entity_ids:

                # Get filenames for this entity; for sequence-level analysis,
                # show only the first image as an exemplar.
                if options.sequence_level_analysis:
                    all_fns = seq_id_to_filenames_map.get(entity_id, [])
                    fns = all_fns[:1]
                else:
                    fns = [entity_id]

                for fn in fns:
                    im = results_fn_to_im.get(fn, None)
                    if im is None:
                        continue

                    fn_clean = flatten_path(fn).replace(' ', '_')
                    image_link = 'images/' + fn_clean

                    gt_cats_str = ', '.join(sorted(filename_to_gt_categories.get(
                        entity_id if not options.sequence_level_analysis else entity_id,
                        set())))
                    pred_cats = filename_to_pred_categories.get(
                        entity_id if not options.sequence_level_analysis else entity_id, {})
                    pred_cats_str = ', '.join(
                        ['{} ({:.2f})'.format(c, conf)
                         for c, conf in sorted(pred_cats.items())])

                    title = '<b>File</b>: {}<br/><b>GT</b>: {}<br/><b>Pred</b>: {}'.format(
                        fn, gt_cats_str, pred_cats_str)

                    html_image_info = {
                        'filename': image_link,
                        'title': title,
                        'textStyle':
                            'font-family:verdana,arial,calibri;font-size:80%;'
                            'text-align:left;margin-top:20;margin-bottom:5'
                    }
                    html_image_info_list.append(html_image_info)

            # ...for each entity

            cell_html_filename = 'predicted_{}_true_{}.html'.format(
                pred_cat.replace(' ', '_').replace('/', '_'),
                true_cat.replace(' ', '_').replace('/', '_'))
            cell_html_path = os.path.join(options.html_output_dir, cell_html_filename)

            cell_title = 'True: {}, Predicted: {}'.format(true_cat, pred_cat)
            cell_options = {
                'headerHtml': '<h1>{}</h1>'.format(cell_title),
                'maxFiguresPerHtmlFile': options.max_images_per_html_file
            }

            write_html_image_list(
                filename=cell_html_path,
                images=html_image_info_list,
                options=cell_options)

        # ...for each cell

        # Generate per-cell HTML pages for misprediction subsets.
        #
        # FP and FN perspectives have separate pages because they can contain
        # different entity sets for the same (true_cat, pred_cat) pair.

        def _fp_cell_html_filename(pred_cat, true_cat):
            return 'fp_predicted_{}_true_{}.html'.format(
                pred_cat.replace(' ', '_').replace('/', '_'),
                true_cat.replace(' ', '_').replace('/', '_'))

        def _fn_cell_html_filename(true_cat, pred_cat):
            return 'fn_true_{}_predicted_{}.html'.format(
                true_cat.replace(' ', '_').replace('/', '_'),
                pred_cat.replace(' ', '_').replace('/', '_'))

        def _build_cell_image_list(entity_ids):
            html_image_info_list = []
            for entity_id in entity_ids:
                if options.sequence_level_analysis:
                    all_fns = seq_id_to_filenames_map.get(entity_id, [])
                    fns = all_fns[:1]
                else:
                    fns = [entity_id]

                for fn in fns:
                    im = results_fn_to_im.get(fn, None)
                    if im is None:
                        continue

                    fn_clean = flatten_path(fn).replace(' ', '_')
                    image_link = 'images/' + fn_clean

                    gt_cats_str = ', '.join(sorted(filename_to_gt_categories.get(
                        entity_id, set())))
                    pred_cats_for_entity = filename_to_pred_categories.get(entity_id, {})
                    pred_cats_str = ', '.join(
                        ['{} ({:.2f})'.format(c, conf)
                         for c, conf in sorted(pred_cats_for_entity.items())])

                    title = '<b>File</b>: {}<br/><b>GT</b>: {}<br/><b>Pred</b>: {}'.format(
                        fn, gt_cats_str, pred_cats_str)

                    html_image_info = {
                        'filename': image_link,
                        'title': title,
                        'textStyle':
                            'font-family:verdana,arial,calibri;font-size:80%;'
                            'text-align:left;margin-top:20;margin-bottom:5'
                    }
                    html_image_info_list.append(html_image_info)
            return html_image_info_list

        # FP pages: "mispredicted as this category"
        for (pred_cat, true_cat), entity_ids in sampled_fp_cells.items():

            html_image_info_list = _build_cell_image_list(entity_ids)
            cell_html_path = os.path.join(
                options.html_output_dir,
                _fp_cell_html_filename(pred_cat, true_cat))
            cell_title = 'Mispredicted as: {} (true: {})'.format(pred_cat, true_cat)

            write_html_image_list(
                filename=cell_html_path,
                images=html_image_info_list,
                options={
                    'headerHtml': '<h1>{}</h1>'.format(cell_title),
                    'maxFiguresPerHtmlFile': options.max_images_per_html_file
                })

        # FN pages: "mispredicted as" (this true category was missed)
        for (true_cat, pred_cat), entity_ids in sampled_fn_cells.items():

            html_image_info_list = _build_cell_image_list(entity_ids)
            cell_html_path = os.path.join(
                options.html_output_dir,
                _fn_cell_html_filename(true_cat, pred_cat))
            cell_title = 'True: {}, mispredicted as: {}'.format(true_cat, pred_cat)

            write_html_image_list(
                filename=cell_html_path,
                images=html_image_info_list,
                options={
                    'headerHtml': '<h1>{}</h1>'.format(cell_title),
                    'maxFiguresPerHtmlFile': options.max_images_per_html_file
                })

        # ...for each misprediction cell

        # Generate index.html

        # Helper to build a cell subpage filename
        def _cell_html_filename(pred_cat, true_cat):
            return 'predicted_{}_true_{}.html'.format(
                pred_cat.replace(' ', '_').replace('/', '_'),
                true_cat.replace(' ', '_').replace('/', '_'))

        # Build per-category misprediction summaries (top-N from off-diagonal)
        n_mispred = options.n_mispredictions_for_table

        # "Mispredicted as this category" = FP column: other true categories
        # predicted as this one (off-diagonal entries in this category's column)
        mispredicted_as = {}

        # "This was mispredicted as" = FN row: this true category predicted as
        # other categories (off-diagonal entries in this category's row)
        this_mispredicted_as = {}

        for cat in active_categories:

            idx = category_to_index[cat]

            # Column: which true categories were predicted as this one?
            col_entries = []
            for other_cat in active_categories:
                if other_cat == cat:
                    continue
                other_idx = category_to_index[other_cat]
                count = int(confusion_matrix[other_idx, idx])
                if count > 0:
                    col_entries.append((other_cat, count))
            col_entries.sort(key=lambda x: -x[1])
            mispredicted_as[cat] = col_entries[:n_mispred]

            # Row: what was this true category predicted as?
            row_entries = []
            for other_cat in active_categories:
                if other_cat == cat:
                    continue
                other_idx = category_to_index[other_cat]
                count = int(confusion_matrix[idx, other_idx])
                if count > 0:
                    row_entries.append((other_cat, count))
            row_entries.sort(key=lambda x: -x[1])
            this_mispredicted_as[cat] = row_entries[:n_mispred]

        # ...for each category

        style_header = """<head>
    <style type="text/css">
    a { text-decoration: none; }
    body { font-family: segoe ui, calibri, "trebuchet ms", verdana, arial, sans-serif; }
    div.contentdiv { margin-left: 20px; }
    table.result-table { border:1px solid black; border-collapse: collapse; margin-left:50px;}
    td,th { padding:10px; border:1px solid #ddd; }
    th.sortable { cursor: pointer; user-select: none; }
    th.sortable:hover { background-color: #e8e8e8; }
    th.sortable::after { content: " \\2195"; font-size: 80%; color: #999; }
    .rotate {
      padding:0px;
      writing-mode:vertical-lr;
      -webkit-transform: rotate(-180deg);
      -moz-transform: rotate(-180deg);
      -ms-transform: rotate(-180deg);
      -o-transform: rotate(-180deg);
      transform: rotate(-180deg);
    }
    .metrics-table td { text-align: right; padding: 5px 15px; }
    .metrics-table td:first-child { text-align: left; font-weight: bold; }
    .mispred-cell { font-size: 85%; line-height: 1.6; }
    </style>
    <script>
    function sortTable(tableId, colIdx, type) {
      var table = document.getElementById(tableId);
      var tbody = table.tBodies[0];
      var rows = Array.from(tbody.rows);
      var asc = table.getAttribute('data-sort-col') == colIdx &&
                table.getAttribute('data-sort-dir') == 'asc' ? false : true;
      rows.sort(function(a, b) {
        var va = a.cells[colIdx].getAttribute('data-val') || a.cells[colIdx].textContent;
        var vb = b.cells[colIdx].getAttribute('data-val') || b.cells[colIdx].textContent;
        if (type === 'num') { va = parseFloat(va) || 0; vb = parseFloat(vb) || 0; }
        else { va = va.toLowerCase(); vb = vb.toLowerCase(); }
        if (va < vb) return asc ? -1 : 1;
        if (va > vb) return asc ? 1 : -1;
        return 0;
      });
      rows.forEach(function(r) { tbody.appendChild(r); });
      table.setAttribute('data-sort-col', colIdx);
      table.setAttribute('data-sort-dir', asc ? 'asc' : 'desc');
    }
    </script>
    </head>"""

        html = '<html>\n'
        html += style_header + '\n'
        html += '<body>\n'

        html += '<h1>Classification Analysis Results</h1>\n'

        # Summary / metadata
        html += '<div class="contentdiv">\n'
        html += '<p><b>Results file</b>: {}</p>\n'.format(os.path.basename(options.results_file))
        html += '<p><b>Ground truth file</b>: {}</p>\n'.format(os.path.basename(options.gt_file))
        html += '<p><b>Detection threshold</b>: {}</p>\n'.format(detection_threshold)
        html += '<p><b>Classification confidence threshold</b>: {}</p>\n'.format(
            options.classification_confidence_threshold)

        if options.sequence_level_analysis:
            html += '<p><b>Analysis level</b>: sequence</p>\n'
        else:
            html += '<p><b>Analysis level</b>: image</p>\n'

        analysis_unit = 'sequences' if options.sequence_level_analysis else 'images'
        html += '<p><b>Total {}</b>: {}</p>\n'.format(analysis_unit,
                                                       len(filename_to_gt_categories))

        if len(categories_to_ignore) > 0:
            html += '<p><b>Excluded categories</b>: {}</p>\n'.format(
                ', '.join(sorted(categories_to_ignore)))

        if options.single_label_per_image:
            html += '<p><b>Ground truth collapsed</b>: single label per {}</p>\n'.format(
                analysis_unit.rstrip('s'))

        if options.single_prediction_per_image:
            html += '<p><b>Predictions collapsed</b>: single label per {}</p>\n'.format(
                analysis_unit.rstrip('s'))

        html += '</div>\n'

        # Overall metrics
        if options.show_overall_metrics:

            html += '<h2>Overall metrics</h2>\n'
            html += '<div class="contentdiv">\n'
            html += '<table class="metrics-table">\n'
            html += '<tr><td>Accuracy</td><td>{:.4f}</td></tr>\n'.format(accuracy)
            html += '<tr><td>Macro F1</td><td>{:.4f}</td></tr>\n'.format(macro_f1)
            html += '<tr><td>Micro F1</td><td>{:.4f}</td></tr>\n'.format(micro_f1)
            html += '<tr><td>Micro Precision</td><td>{:.4f}</td></tr>\n'.format(micro_precision)
            html += '<tr><td>Micro Recall</td><td>{:.4f}</td></tr>\n'.format(micro_recall)
            html += '</table>\n'
            html += '</div>\n'

        # Column explanations
        html += '<h2>Column explanations</h2>\n'
        html += '<div class="contentdiv">\n'
        html += '<p><b>Predicted as this category</b>: images where this category was among the predictions, '
        html += 'and some other category was among the true labels. This category may also be a correct true '
        html += 'label for the image; these entries do not necessarily represent errors. For example, if the '
        html += 'ground truth for an image is &ldquo;human,cattle&rdquo; and the prediction is &ldquo;human&rdquo; '
        html += 'or &ldquo;human,cattle&rdquo;, this image would be included in the "human" link in this column in '
        html += 'the &ldquo;cattle&rdquo; row. This column is '
        html += 'generally useful for understanding the distribution of multi-label or multi-prediction images, '
        html += 'but it is not very useful for understanding model errors.</p>\n'
        html += '<p><b>Mispredicted as this category</b>: images where this category was among the predictions, '
        html += 'and it was wrong (not present in the ground truth). This column is useful for understanding '
        html += 'model errors, specifically false positives for this category.</p>\n'
        html += '<p><b>This was predicted as</b>: images with this true label that also received a prediction '
        html += 'of some other category. Other predictions (including the correct one) may also be present. '
        html += 'For example, if the ground truth for an image is &ldquo;cattle&rdquo;, and the prediction is '
        html += '&ldquo;human,cattle&rdquo;, this image would be included in the "human" link in this column in '
        html += 'the &ldquo;cattle&rdquo; row. This column is generally '
        html += 'useful for understanding the distribution of multi-label or multi-prediction images, but it is '
        html += 'not very useful for understanding model errors.</p>\n'
        html += '<p><b>Mispredicted as</b>: images with this true label whose predictions did not include this '
        html += 'label. This column is useful for understanding model errors, specifically false negatives for '
        html += 'this category.</p>\n'
        html += '</div>\n'

        # Per-category statistics table (sortable)
        html += '<h2>Per-category statistics</h2>\n'
        html += '<p style="margin-left:50px;font-size:90%;">Click column headers to sort</p>\n'
        html += '<table class="result-table" id="stats-table">\n'
        html += '<thead><tr>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',0,\'str\')">Category</th>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',1,\'num\')">N (GT)</th>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',2,\'num\')">N (Pred)</th>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',3,\'num\')">Precision</th>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',4,\'num\')">Recall</th>'
        html += '<th class="sortable" onclick="sortTable(\'stats-table\',5,\'num\')">F1</th>'
        html += '<th>Predicted as this category</th>'
        html += '<th>Mispredicted as this category</th>'
        html += '<th>This was predicted as</th>'
        html += '<th>Mispredicted as</th>'
        html += '</tr></thead>\n'
        html += '<tbody>\n'

        for cat in active_categories:

            r = per_category_results[cat]
            html += '<tr>'
            html += '<td>{}</td>'.format(cat)
            html += '<td data-val="{}">{}</td>'.format(r['n_ground_truth'], r['n_ground_truth'])
            html += '<td data-val="{}">{}</td>'.format(r['n_predicted'], r['n_predicted'])
            html += '<td data-val="{:.6f}">{:.3f}</td>'.format(r['precision'], r['precision'])
            html += '<td data-val="{:.6f}">{:.3f}</td>'.format(r['recall'], r['recall'])
            html += '<td data-val="{:.6f}">{:.3f}</td>'.format(r['f1'], r['f1'])

            # "Mispredicted as this category" column (FP: other things predicted as cat)
            fp_entries = mispredicted_as[cat]
            if len(fp_entries) == 0:
                html += '<td></td>'
            else:
                html += '<td class="mispred-cell">'
                fp_parts = []
                for other_cat, count in fp_entries:
                    link = _cell_html_filename(cat, other_cat)
                    fp_parts.append('<a href="{}">{}</a> ({})'.format(
                        link, other_cat, count))
                html += '<br/>'.join(fp_parts)
                html += '</td>'

            # "Mispredicted as this category" column (FP: this category was
            # predicted but is not in GT)
            mispred_fp_entries = mispredicted_as_this[cat]
            if len(mispred_fp_entries) == 0:
                html += '<td></td>'
            else:
                html += '<td class="mispred-cell">'
                mispred_fp_parts = []
                for other_cat, count in mispred_fp_entries:
                    link = _fp_cell_html_filename(cat, other_cat)
                    mispred_fp_parts.append('<a href="{}">{}</a> ({})'.format(
                        link, other_cat, count))
                html += '<br/>'.join(mispred_fp_parts)
                html += '</td>'

            # "This was predicted as" column (off-diagonal confusion matrix row)
            fn_entries = this_mispredicted_as[cat]
            if len(fn_entries) == 0:
                html += '<td></td>'
            else:
                html += '<td class="mispred-cell">'
                fn_parts = []
                for other_cat, count in fn_entries:
                    link = _cell_html_filename(other_cat, cat)
                    fn_parts.append('<a href="{}">{}</a> ({})'.format(
                        link, other_cat, count))
                html += '<br/>'.join(fn_parts)
                html += '</td>'

            # "Mispredicted as" column (FN: this true category was not predicted)
            mispred_fn_entries = this_was_mispredicted_as[cat]
            if len(mispred_fn_entries) == 0:
                html += '<td></td>'
            else:
                html += '<td class="mispred-cell">'
                mispred_fn_parts = []
                for other_cat, count in mispred_fn_entries:
                    link = _fn_cell_html_filename(cat, other_cat)
                    mispred_fn_parts.append('<a href="{}">{}</a> ({})'.format(
                        link, other_cat, count))
                html += '<br/>'.join(mispred_fn_parts)
                html += '</td>'

            html += '</tr>\n'

        # ...for each category

        html += '</tbody>\n'
        html += '</table>\n'

        # Confusion matrix
        html += '<h2>Confusion matrix</h2>\n'
        html += '<p>Rows = true categories, columns = predicted categories. '
        html += 'On-diagonal elements indicate a category that was both in the ground truth and among '
        html += 'the predictions. An image with multiple correct labels contributes to the diagonal for '
        html += 'each correct category. The off-diagonal elements do not necessarily indicate errors; they '
        html += 'correspond to the &ldquo;predicted as this category&rdquo; and &ldquo;this was predicted '
        html += 'as&rdquo; columns in the table above.</p>\n'
        html += '<table class="result-table">\n'

        # Header row with rotated predicted category labels
        html += '<tr><td>&nbsp;</td>\n'
        for cat in active_categories:
            html += '<td class="rotate"><p style="margin-left:20px;">{}</p></td>\n'.format(cat)
        html += '</tr>\n'

        # Data rows
        for true_cat in active_categories:

            html += '<tr>\n'
            html += '<td><b>{}</b></td>\n'.format(true_cat)

            for pred_cat in active_categories:

                true_idx = category_to_index[true_cat]
                pred_idx = category_to_index[pred_cat]
                count = int(confusion_matrix[true_idx, pred_idx])

                if count == 0:
                    html += '<td></td>\n'
                else:
                    cell_filename = _cell_html_filename(pred_cat, true_cat)

                    # Highlight diagonal (correct predictions) with background color
                    if true_cat == pred_cat:
                        html += '<td style="background-color:#d4edda;">'
                    else:
                        html += '<td>'

                    html += '<a href="{}">{}</a></td>\n'.format(cell_filename, count)

            # ...for each predicted category

            html += '</tr>\n'

        # ...for each true category

        html += '</table>\n'

        # Default sort: N (GT) descending (call twice to toggle from asc to desc)
        html += '<script>sortTable("stats-table",1,"num");sortTable("stats-table",1,"num");</script>\n'

        html += '</body>\n'
        html += '</html>\n'

        html_output_file = os.path.join(options.html_output_dir, 'index.html')
        with open(html_output_file, 'w') as f:
            f.write(html)

        print('\nHTML output written to {}'.format(html_output_file))

    # ...if we're supposed to write HTML output

    ## Build and return results

    results = AnalysisResults()
    results.per_category_results = per_category_results
    results.macro_f1 = float(macro_f1)
    results.micro_f1 = float(micro_f1)
    results.micro_precision = float(micro_precision)
    results.micro_recall = float(micro_recall)
    results.accuracy = float(accuracy)
    results.confusion_matrix = confusion_matrix
    results.active_categories = active_categories
    results.html_output_file = html_output_file

    return results

# ...def analyze_classification_results(...)


def render_misprediction_pages(options, cells_to_render):
    """
    Render detailed HTML pages for specific misprediction cells, typically with a
    large number of images (e.g. 2000) for deep-dive analysis.

    Uses the same data-loading and preparation logic as analyze_classification_results.

    Args:
        options (ClassificationAnalysisOptions): options object; needs results_file,
            gt_file, image_base_dir, html_output_dir.  max_images_per_cell controls
            page length (e.g. set to 2000 for deep dives).
        cells_to_render (list): list of (true_cat, pred_cat, mode) tuples where mode
            is one of:

            - 'standard': entities where true_cat is in GT, pred_cat is in predictions,
              true_cat != pred_cat, and the off-diagonal skip condition is not met.
            - 'strict_fp': entities where pred_cat is among the predictions and
              pred_cat is NOT in the ground truth, and true_cat is in GT.  (The
              "mispredicted as this category" criterion.)
            - 'strict_fn': entities where true_cat is in GT and true_cat is NOT
              among the predictions, and pred_cat is in predictions.  (The
              "mispredicted as" criterion.)

    Returns:
        list: paths to the generated HTML files
    """

    assert options.image_base_dir is not None, \
        'image_base_dir is required for render_misprediction_pages'
    assert options.html_output_dir is not None, \
        'html_output_dir is required for render_misprediction_pages'

    prepared = _prepare_analysis_data(options)

    filename_to_gt_categories = prepared['filename_to_gt_categories']
    filename_to_pred_categories = prepared['filename_to_pred_categories']
    active_categories = prepared['active_categories']
    results_fn_to_im = prepared['results_fn_to_im']
    gt_data = prepared['gt_data']
    detection_category_id_to_name = prepared['detection_category_id_to_name']
    classification_category_id_to_name = prepared['classification_category_id_to_name']
    detection_threshold = prepared['detection_threshold']

    os.makedirs(options.html_output_dir, exist_ok=True)
    preview_images_folder = os.path.join(options.html_output_dir, 'images')
    os.makedirs(preview_images_folder, exist_ok=True)

    # For sequence-level analysis, build seq_id -> filenames map
    seq_id_to_filenames_map = None
    if options.sequence_level_analysis:
        seq_id_to_filenames_map = defaultdict(list)
        for im in gt_data['images']:
            fn = im['file_name']
            if fn in results_fn_to_im and 'seq_id' in im:
                seq_id_to_filenames_map[im['seq_id']].append(fn)

    # Validate requested categories
    active_set = set(active_categories)
    for true_cat, pred_cat, mode in cells_to_render:
        if true_cat not in active_set:
            print('Warning: true category "{}" not in active categories, '
                  'will produce empty page'.format(true_cat))
        if pred_cat not in active_set:
            print('Warning: pred category "{}" not in active categories, '
                  'will produce empty page'.format(pred_cat))
        assert mode in ('standard', 'strict_fp', 'strict_fn'), \
            'mode must be "standard", "strict_fp", or "strict_fn", got "{}"'.format(mode)

    # Collect matching entity IDs for each requested cell
    cell_entity_ids = {}

    for true_cat, pred_cat, mode in cells_to_render:
        key = (true_cat, pred_cat, mode)
        matching = []

        for entity_id in filename_to_gt_categories:
            gt_cats = filename_to_gt_categories[entity_id]
            pred_cats = filename_to_pred_categories[entity_id]

            if true_cat not in gt_cats:
                continue

            if mode == 'standard':
                if pred_cat not in pred_cats:
                    continue
                if true_cat == pred_cat:
                    continue
                # Off-diagonal skip: both categories are correctly present
                if pred_cat in gt_cats and true_cat in pred_cats:
                    continue
                matching.append(entity_id)

            elif mode == 'strict_fp':
                # pred_cat was predicted and is NOT in GT
                if pred_cat not in pred_cats:
                    continue
                if pred_cat in gt_cats:
                    continue
                matching.append(entity_id)

            else:  # strict_fn
                # true_cat is in GT and is NOT among predictions
                if true_cat in pred_cats:
                    continue
                if pred_cat not in pred_cats:
                    continue
                matching.append(entity_id)

        cell_entity_ids[key] = matching

    # Sample to max_images_per_cell
    rng = random.Random(options.random_seed)
    for key in cell_entity_ids:
        entity_ids = cell_entity_ids[key]
        if len(entity_ids) > options.max_images_per_cell:
            cell_entity_ids[key] = rng.sample(entity_ids, options.max_images_per_cell)

    # Collect all filenames that need rendering
    filenames_to_render = set()
    for entity_ids in cell_entity_ids.values():
        if options.sequence_level_analysis:
            for seq_id in entity_ids:
                fns = seq_id_to_filenames_map.get(seq_id, [])
                if len(fns) > 0:
                    filenames_to_render.add(fns[0])
        else:
            filenames_to_render.update(entity_ids)

    # Build image entries for rendering
    images_to_render = []
    for fn in filenames_to_render:
        if fn in results_fn_to_im:
            images_to_render.append(results_fn_to_im[fn])

    print('\nRendering {} images for {} misprediction pages...'.format(
        len(images_to_render), len(cells_to_render)))

    # Render images
    render_constants = {
        'image_base_dir': options.image_base_dir,
        'output_dir': preview_images_folder,
        'detection_label_map': detection_category_id_to_name,
        'classification_label_map': classification_category_id_to_name,
        'detection_threshold': detection_threshold,
        'classification_confidence_threshold': options.classification_confidence_threshold,
        'output_image_width': options.output_image_width,
        'overwrite': options.overwrite
    }

    if options.rendering_workers > 1 and len(images_to_render) > 1:

        if options.rendering_pool_type == 'threads':
            pool = ThreadPool(options.rendering_workers)
            worker_string = 'threads'
        else:
            pool = Pool(options.rendering_workers)
            worker_string = 'processes'

        print('Rendering with {} {}'.format(options.rendering_workers, worker_string))

        try:
            rendering_results = list(tqdm(
                pool.imap(
                    partial(_render_single_image, render_constants=render_constants),
                    images_to_render),
                total=len(images_to_render)))
        finally:
            pool.close()
            pool.join()
    else:
        rendering_results = []
        for im in tqdm(images_to_render):
            rendering_results.append(_render_single_image(im, render_constants))

    n_success = sum(1 for r in rendering_results if r['success'])
    print('Successfully rendered {} of {} images'.format(n_success, len(images_to_render)))

    # Generate one HTML page per requested cell
    generated_files = []

    for true_cat, pred_cat, mode in cells_to_render:
        key = (true_cat, pred_cat, mode)
        entity_ids = cell_entity_ids[key]

        html_image_info_list = []

        for entity_id in entity_ids:

            if options.sequence_level_analysis:
                all_fns = seq_id_to_filenames_map.get(entity_id, [])
                fns = all_fns[:1]
            else:
                fns = [entity_id]

            for fn in fns:
                im = results_fn_to_im.get(fn, None)
                if im is None:
                    continue

                fn_clean = flatten_path(fn).replace(' ', '_')
                image_link = 'images/' + fn_clean

                gt_cats_str = ', '.join(sorted(filename_to_gt_categories.get(
                    entity_id, set())))
                pred_cats_for_entity = filename_to_pred_categories.get(entity_id, {})
                pred_cats_str = ', '.join(
                    ['{} ({:.2f})'.format(c, conf)
                     for c, conf in sorted(pred_cats_for_entity.items())])

                title = '<b>File</b>: {}<br/><b>GT</b>: {}<br/><b>Pred</b>: {}'.format(
                    fn, gt_cats_str, pred_cats_str)

                html_image_info = {
                    'filename': image_link,
                    'title': title,
                    'textStyle':
                        'font-family:verdana,arial,calibri;font-size:80%;'
                        'text-align:left;margin-top:20;margin-bottom:5'
                }
                html_image_info_list.append(html_image_info)

        # Build filename
        mode_prefix = 'strict_' if mode == 'strict' else ''
        cell_html_filename = '{}predicted_{}_true_{}.html'.format(
            mode_prefix,
            pred_cat.replace(' ', '_').replace('/', '_'),
            true_cat.replace(' ', '_').replace('/', '_'))
        cell_html_path = os.path.join(options.html_output_dir, cell_html_filename)

        mode_label = 'Strictly predicted' if mode == 'strict' else 'Predicted'
        cell_title = '{}: {} (true: {}) — {} images'.format(
            mode_label, pred_cat, true_cat, len(html_image_info_list))
        cell_options = {
            'headerHtml': '<h1>{}</h1>'.format(cell_title),
            'maxFiguresPerHtmlFile': options.max_images_per_html_file
        }

        write_html_image_list(
            filename=cell_html_path,
            images=html_image_info_list,
            options=cell_options)

        generated_files.append(cell_html_path)
        print('  Wrote {} ({} images)'.format(cell_html_filename, len(html_image_info_list)))

    return generated_files

# ...def render_misprediction_pages(...)


#%% Interactive driver

if False:

    #%%

    options = ClassificationAnalysisOptions()

    options.results_file = '/home/user/tmp/classification-analysis/subset_results.json'
    options.gt_file = '/home/user/tmp/classification-analysis/subset_gt.json'
    options.image_base_dir = '/home/user/tmp/images'
    options.html_output_dir = '/home/user/tmp/classification-analysis/html_output'
    options.sequence_level_analysis = False
    options.rendering_pool_type = 'processes'
    options.categories_to_ignore = ['vehicle','no cv result','setup_pickup']
    options.single_prediction_per_image = True
    options.single_label_per_image = True

    results = analyze_classification_results(options)

    from megadetector.utils.path_utils import open_file
    open_file(results.html_output_file)


    #%% Deep-dive into specific misprediction cells

    from megadetector.postprocessing.analyze_classification_results import \
        ClassificationAnalysisOptions, render_misprediction_pages
    from megadetector.utils.path_utils import open_file

    deep_dive_options = ClassificationAnalysisOptions()
    deep_dive_options.results_file = '/home/user/tmp/classification-analysis/subset_results.json'
    deep_dive_options.gt_file = '/home/user/tmp/classification-analysis/subset_gt.json'
    deep_dive_options.image_base_dir = '/home/user/tmp/images'
    deep_dive_options.html_output_dir = '/home/user/tmp/classification-analysis/deep_dive'
    deep_dive_options.sequence_level_analysis = False
    deep_dive_options.categories_to_ignore = ['vehicle','no cv result','setup_pickup']
    deep_dive_options.single_prediction_per_image = False
    deep_dive_options.single_label_per_image = False
    deep_dive_options.max_images_per_cell = 2000
    deep_dive_options.rendering_pool_type = 'processes'

    cells = [
        ('domestic cattle', 'empty', 'strict_fp'),
        ('domestic cattle', 'human', 'strict_fn'),
    ]

    generated_files = render_misprediction_pages(deep_dive_options, cells)
    for f in generated_files:
        open_file(f)


#%% Command-line driver

def main():
    """
    Command-line driver for analyze_classification_results
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Analyze classification results against ground truth, '
                    'computing precision/recall/F1 and generating an HTML report.')

    parser.add_argument(
        'results_file', type=str,
        help='MD-formatted results file (.json)')

    parser.add_argument(
        'gt_file', type=str,
        help='Ground truth file in COCO Camera Traps format (.json)')

    parser.add_argument(
        '--image_base_dir', type=str, default=None,
        help='Folder where images live; required if --html_output_dir is specified')

    parser.add_argument(
        '--html_output_dir', type=str, default=None,
        help='Folder for HTML output with confusion matrix and image galleries')

    parser.add_argument(
        '--detection_threshold', type=float, default=None,
        help='Detection confidence threshold (auto-detected if not specified)')

    parser.add_argument(
        '--classification_confidence_threshold', type=float, default=0.5,
        help='Classification confidence threshold')

    parser.add_argument(
        '--max_total_images', type=int, default=8000,
        help='Maximum total number of images to render')

    parser.add_argument(
        '--max_images_per_cell', type=int, default=50,
        help='Maximum number of images per confusion matrix cell')

    parser.add_argument(
        '--random_seed', type=int, default=0,
        help='Random seed for image sampling')

    parser.add_argument(
        '--sequence_level', action='store_true',
        help='Perform analysis at the sequence level instead of image level')

    parser.add_argument(
        '--rendering_workers', type=int, default=10,
        help='Number of workers for image rendering')

    parser.add_argument(
        '--rendering_pool_type', type=str, default='threads',
        choices=['threads', 'processes'],
        help='Type of worker pool for rendering')

    parser.add_argument(
        '--output_image_width', type=int, default=1000,
        help='Width of rendered output images (-1 for original size)')

    parser.add_argument(
        '--n_mispredictions_for_table', type=int, default=3,
        help='Number of top misprediction categories to show in the per-category table')

    parser.add_argument(
        '--categories_to_ignore', type=str, default=None,
        help='Comma-separated list of category names to exclude from analysis')

    parser.add_argument(
        '--single_prediction_per_image', action='store_true',
        help='Collapse predictions to a single category per image/sequence')

    parser.add_argument(
        '--single_label_per_image', action='store_true',
        help='Collapse ground truth to a single category per image/sequence')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    options = ClassificationAnalysisOptions()
    options.results_file = args.results_file
    options.gt_file = args.gt_file
    options.image_base_dir = args.image_base_dir
    options.html_output_dir = args.html_output_dir
    options.detection_threshold = args.detection_threshold
    options.classification_confidence_threshold = args.classification_confidence_threshold
    options.max_total_images = args.max_total_images
    options.max_images_per_cell = args.max_images_per_cell
    options.random_seed = args.random_seed
    options.sequence_level_analysis = args.sequence_level
    options.rendering_workers = args.rendering_workers
    options.rendering_pool_type = args.rendering_pool_type
    options.output_image_width = args.output_image_width
    options.n_mispredictions_for_table = args.n_mispredictions_for_table
    if args.categories_to_ignore is not None:
        options.categories_to_ignore = [c.strip() for c in args.categories_to_ignore.split(',')]
    options.single_prediction_per_image = args.single_prediction_per_image
    options.single_label_per_image = args.single_label_per_image

    results = analyze_classification_results(options)

    print('\nResults:')
    print('  Accuracy: {:.4f}'.format(results.accuracy))
    print('  Macro F1: {:.4f}'.format(results.macro_f1))
    print('  Micro F1: {:.4f}'.format(results.micro_f1))

    return results


if __name__ == '__main__':
    main()



r"""

separate_detections_into_folders.py

**Overview**

Given a .json file with batch processing results, separate the files in that
set of results into folders that contain animals/people/vehicles/nothing,
according to per-class thresholds.

Image files are copied, not moved.

**Output structure**

Preserves relative paths within each of those folders; cannot be used with .json
files that have absolute paths in them.

For example, if your .json file has these images:

* a/b/c/1.jpg
* a/b/d/2.jpg
* a/b/e/3.jpg
* a/b/f/4.jpg
* a/x/y/5.jpg

And let's say:

* The results say that the first three images are empty/person/vehicle, respectively
* The fourth image is above threshold for "animal" and "person"
* The fifth image contains an animal

* You specify an output base folder of c:/out

You will get the following files:

* c:/out/empty/a/b/c/1.jpg
* c:/out/people/a/b/d/2.jpg
* c:/out/vehicles/a/b/e/3.jpg
* c:/out/animal_person/a/b/f/4.jpg
* c:/out/animals/a/x/y/5.jpg

**Rendering bounding boxes**

By default, images are just copied to the target output folder.  If you specify --render_boxes,
bounding boxes will be rendered on the output images.  Because this is no longer strictly
a copy operation, this may result in the loss of metadata.  More accurately, this *may*
result in the loss of some EXIF metadata; this *will* result in the loss of IPTC/XMP metadata.

Rendering boxes also makes this script a lot slower.

**Classification-based separation**

If you have a results file with classification data, you can also specify classes to put
in their own folders, within the "animals" folder, like this:

``--classification_thresholds "deer=0.75,cow=0.75"``

So, e.g., you might get:

c:/out/animals/deer/a/x/y/5.jpg

In this scenario, the folders within "animals" will be:

deer, cow, multiple, unclassified

"multiple" in this case only means "deer and cow"; if an image is classified as containing a
bird and a bear, that would end up in "unclassified", since the folder separation is based only
on the categories you provide at the command line.

No classification-based separation is done within the animal_person, animal_vehicle, or
animal_person_vehicle folders.

"""

#%% Constants and imports

import argparse
import json
import os
import shutil
import sys
import itertools

from multiprocessing.pool import ThreadPool
from functools import partial
from tqdm import tqdm

from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.string_utils import is_float
from megadetector.utils.path_utils import remove_empty_folders
from megadetector.detection.run_detector import get_typical_confidence_threshold_from_results
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.visualization.visualization_utils import blur_detections

friendly_folder_names = {'animal':'animals','person':'people','vehicle':'vehicles'}

# Occasionally we have near-zero confidence detections associated with COCO classes that
# didn't quite get squeezed out of the model in training.  As long as they're near zero
# confidence, we just ignore them.
invalid_category_epsilon = 0.00001

default_line_thickness = 8
default_box_expansion = 3


#%% Options class

class SeparateDetectionsIntoFoldersOptions:
    """
    Options used to parameterize separate_detections_into_folders()
    """

    def __init__(self,threshold=None):

        #: Default threshold for categories not specified in category_name_to_threshold
        self.threshold = None

        #: Dict mapping category names to thresholds; for example, an image with only a detection of class
        #: "animal" whose confidence is greater than or equal to category_name_to_threshold['animal']
        #: will be put in the "animal" folder.
        self.category_name_to_threshold = {
            'animal': self.threshold,
            'person': self.threshold,
            'vehicle': self.threshold
        }

        #: Number of workers to use, set to <= 1 to disable parallelization
        self.n_threads = 1

        #: By default, this function errors if you try to output to an existing folder
        self.allow_existing_directory = False

        #: By default, this function errors if any of the images specified in the results file don't
        #: exist in the source folder.
        self.allow_missing_files = False

        #: Whether to overwrite images that already exist in the target folder; only relevant if
        #: [allow_existing_directory] is True
        self.overwrite = True

        #: Whether to skip empty images; if this is False, empty images (i.e., images with no detections
        #: above the corresponding threshold) will be copied to an "empty" folder.
        self.skip_empty_images = False

        #: The MD results .json file to process
        self.results_file = None

        #: The folder containing source images; filenames in [results_file] should be relative to this
        #: folder.
        self.base_input_folder = None

        #: The folder to which we should write output images; see the module header comment for information
        #: about how that folder will be structured.
        self.base_output_folder = None

        #: Should we move rather than copy?
        self.move_images = False

        #: Should we render boxes on the output images?  Makes everything a lot slower.
        self.render_boxes = False

        #: Line thickness in pixels; only relevant if [render_boxes] is True
        self.line_thickness = default_line_thickness

        #: Box expansion in pixels; only relevant if [render_boxes] is True
        self.box_expansion = default_box_expansion

        #: Originally specified as a string that looks like this:
        #:
        #: deer=0.75,cow=0.75
        #:
        #: String, converted internally to a dict mapping name:threshold
        self.classification_thresholds = None

        ## Debug or internal attributes

        #: Do not set explicitly; populated from data when using classification results
        self.classification_category_id_to_name = None

        #: Do not set explicitly; populated from data when using classification results
        self.classification_categories = None

        #: Used to test this script; sets a limit on the number of images to process.
        self.debug_max_images = None

        #: Do not set explicitly; this gets created based on [results_file]
        #:
        #:Dictionary mapping categories (plus combinations of categories, and 'empty') to output folders
        self.category_name_to_folder = None

        #: Do not set explicitly; this gets loaded from [results_file]
        self.category_id_to_category_name = None

        #: List of category names for which we should blur detections, most commonly ['person']
        #:
        #: Can also be a comma-separated list.
        self.category_names_to_blur = None

        #: Remove all empty folders from the target folder at the end of the process,
        #: whether or not they were created by this script
        self.remove_empty_folders = False

    # ...__init__()

# ...class SeparateDetectionsIntoFoldersOptions


#%% Support functions

def _path_is_abs(p): return (len(p) > 1) and (p[0] == '/' or p[1] == ':')

printed_missing_file_warning = False

def _process_detections(im,options):
    """
    Process all detections for a single image

    May modify *im*.
    """

    global printed_missing_file_warning

    relative_filename = im['file']

    detections = None
    if 'detections' in im:
        detections = im['detections']

    categories_above_threshold = None

    if detections is None:

        assert im['failure'] is not None and len(im['failure']) > 0
        target_folder = options.category_name_to_folder['failure']

    else:

        category_name_to_max_confidence = {}
        category_names = options.category_id_to_category_name.values()
        for category_name in category_names:
            category_name_to_max_confidence[category_name] = 0.0

        # Find the maximum confidence for each category
        #
        # det = detections[0]
        for det in detections:

            category_id = det['category']

            # For zero-confidence detections, we occasionally have leftover goop
            # from COCO classes
            if category_id not in options.category_id_to_category_name:
                print('Warning: unrecognized category {} in file {}'.format(
                    category_id,relative_filename))
                # assert det['conf'] < invalid_category_epsilon
                continue

            category_name = options.category_id_to_category_name[category_id]
            if det['conf'] > category_name_to_max_confidence[category_name]:
                category_name_to_max_confidence[category_name] = det['conf']

        # ...for each detection on this image

        # Count the number of thresholds exceeded
        categories_above_threshold = []
        for category_name in category_names:

            threshold = options.category_name_to_threshold[category_name]
            assert threshold is not None

            max_confidence_this_category = category_name_to_max_confidence[category_name]
            if max_confidence_this_category >= threshold:
                categories_above_threshold.append(category_name)

        # ...for each category

        categories_above_threshold.sort()

        using_classification_folders = (options.classification_thresholds is not None and \
                                        len(options.classification_thresholds) > 0)

        # If this is above multiple thresholds
        if len(categories_above_threshold) > 1:

            # Currently "animal_person" images get put into the "animal_person" folder, even if we're
            # doing species-based separation.  Ideally, we would optionally put these in either the "deer"
            # folder or a "deer_person" folder, but this is pretty esoteric, so not worrying about this
            # for now.
            target_folder = options.category_name_to_folder['_'.join(categories_above_threshold)]

        elif len(categories_above_threshold) == 0:

            target_folder = options.category_name_to_folder['empty']

        else:

            assert len(categories_above_threshold) == 1

            target_folder = options.category_name_to_folder[categories_above_threshold[0]]

            # Are we making species classification folders, and is this an animal?
            if ('animal' in categories_above_threshold) and (using_classification_folders):

                # Do we need to put this into a specific species folder?

                # Find the animal-class detections that are above threshold
                category_name_to_id = {v: k for k, v in options.category_id_to_category_name.items()}
                animal_category_id = category_name_to_id['animal']
                valid_animal_detections = [d for d in detections if \
                                           (d['category'] == animal_category_id and \
                                           d['conf'] >= options.category_name_to_threshold['animal'])]

                # Count the number of classification categories that are above threshold for at
                # least one detection
                classification_categories_above_threshold = set()

                # d = valid_animal_detections[0]
                for d in valid_animal_detections:

                    if 'classifications' not in d or d['classifications'] is None:
                        continue

                    # classification = d['classifications'][0]
                    for classification in d['classifications']:

                        classification_category_id = classification[0]
                        classification_confidence = classification[1]

                        # Do we have a threshold for this category, and if so, is
                        # this classification above threshold?
                        assert options.classification_category_id_to_name is not None
                        classification_category_name = \
                            options.classification_category_id_to_name[classification_category_id]
                        if (classification_category_name in options.classification_thresholds) and \
                            (classification_confidence > \
                             options.classification_thresholds[classification_category_name]):
                            classification_categories_above_threshold.add(classification_category_name)

                    # ...for each classification

                # ...for each detection

                if len(classification_categories_above_threshold) == 0:
                    classification_folder_name = 'unclassified'

                elif len(classification_categories_above_threshold) > 1:
                    classification_folder_name = 'multiple'

                else:
                    assert len(classification_categories_above_threshold) == 1
                    classification_folder_name = list(classification_categories_above_threshold)[0]

                target_folder = os.path.join(target_folder,classification_folder_name)

            # ...if we have to deal with classification subfolders

        # ...if we have 0/1/more categories above threshold

    # ...if this is/isn't a failure case

    source_path = os.path.join(options.base_input_folder,relative_filename)
    if not os.path.isfile(source_path):
        if not options.allow_missing_files:
            raise ValueError('Cannot find file {}'.format(source_path))
        else:
            if not printed_missing_file_warning:
                print('Warning: cannot find at least one file ({})'.format(source_path))
                printed_missing_file_warning = True
            return

    target_path = os.path.join(target_folder,relative_filename)
    if (not options.overwrite) and (os.path.isfile(target_path)):
        return

    target_dir = os.path.dirname(target_path)
    os.makedirs(target_dir,exist_ok=True)

    # Skip this image if it's empty and we're not processing empty images
    if ((categories_above_threshold is None) or (len(categories_above_threshold) == 0)) and \
        options.skip_empty_images:
        return

    # At this point, this image is getting copied; we may or may not also need to
    # draw bounding boxes or blur pixels.

    # Do a simple copy operation if we don't need to manipulate the images (render boxes, blur pixels)
    if (not options.render_boxes and (options.category_names_to_blur is None)) or \
        (categories_above_threshold is None) or \
        (len(categories_above_threshold) == 0):

        if options.move_images:
            shutil.move(source_path,target_path)
        else:
            shutil.copyfile(source_path,target_path)

    else:

        # Open the source image
        pil_image = vis_utils.load_image(source_path)

        # Blur regions in the image if necessary
        category_names_to_blur = options.category_names_to_blur

        if category_names_to_blur is not None:

            if isinstance(category_names_to_blur,str):
                category_names_to_blur = category_names_to_blur.split(',')
                category_names_to_blur = [s.strip() for s in category_names_to_blur]

            detections_to_blur = []
            for d in detections:
                category_name = options.category_id_to_category_name[d['category']]
                category_threshold = options.category_name_to_threshold[category_name]
                if (d['conf'] >= category_threshold) and (category_name in category_names_to_blur):
                    detections_to_blur.append(d)
            if len(detections_to_blur) > 0:
                blur_detections(pil_image,detections_to_blur)

        # Render bounding boxes for each category separately, because
        # we allow different thresholds for each category.

        category_name_to_id = {v: k for k, v in options.category_id_to_category_name.items()}
        assert len(category_name_to_id) == len(options.category_id_to_category_name)

        classification_label_map = None
        if using_classification_folders:
            classification_label_map = options.classification_categories

        for category_name in categories_above_threshold:

            category_id = category_name_to_id[category_name]
            category_threshold = options.category_name_to_threshold[category_name]
            assert category_threshold is not None
            category_detections = [d for d in detections if d['category'] == category_id]

            # When we're not using classification folders, remove classification
            # information to maintain standard detection colors.
            if not using_classification_folders:
                for d in category_detections:
                    if 'classifications' in d:
                        del d['classifications']

            vis_utils.render_detection_bounding_boxes(
                category_detections,
                pil_image,
                label_map=options.detection_categories,
                classification_label_map=classification_label_map,
                confidence_threshold=category_threshold,
                thickness=options.line_thickness,
                expansion=options.box_expansion)

        # ...for each category

        # Try to preserve EXIF data and image quality when saving
        vis_utils.exif_preserving_save(pil_image,target_path)

    # ...if we don't/do need to render boxes

# ...def _process_detections()


#%% Main function

def separate_detections_into_folders(options):
    """
    Given a .json file with batch processing results, separate the files in that
    set of results into folders that contain animals/people/vehicles/nothing,
    according to per-class thresholds.  See the header comment of this module for
    more details about the output folder structure.

    Args:
        options (SeparateDetectionsIntoFoldersOptions): parameters guiding image
        separation, see the SeparateDetectionsIntoFoldersOptions documentation for specific
        options.
    """

    # Input validation

    # Currently we don't support moving (instead of copying) when we're also rendering
    # bounding boxes or blurring humans.
    assert not (options.render_boxes and options.move_images), \
        'Cannot specify both render_boxes and move_images'
    assert not ((options.category_names_to_blur is not None) and options.move_images), \
        'Cannot specify both category_names_to_blur and move_images'

    # Create output folder if necessary
    if (os.path.isdir(options.base_output_folder)) and \
        (len(os.listdir(options.base_output_folder) ) > 0):
        if options.allow_existing_directory:
            print('Warning: target folder exists and is not empty... did ' + \
                  'you mean to delete an old version?')
        else:
            raise ValueError('Target folder exists and is not empty')
    os.makedirs(options.base_output_folder,exist_ok=True)

    # Load detection results
    print('Loading detection results')
    with open(options.results_file,'r') as f:
        results = json.load(f)
    images = results['images']

    for im in images:
        fn = im['file']
        assert not _path_is_abs(fn), 'Cannot process results with absolute image paths'

    print('Processing detections for {} images'.format(len(images)))

    default_threshold = options.threshold

    if default_threshold is None:
        default_threshold = get_typical_confidence_threshold_from_results(results)

    detection_categories = results['detection_categories']
    options.detection_categories = detection_categories
    options.category_id_to_category_name = detection_categories

    # Map class names to output folders
    options.category_name_to_folder = {}
    options.category_name_to_folder['empty'] = os.path.join(options.base_output_folder,'empty')
    options.category_name_to_folder['failure'] =\
        os.path.join(options.base_output_folder,'processing_failure')

    # Create all combinations of categories
    category_names = list(detection_categories.values())
    category_names.sort()

    # category_name = category_names[0]
    for category_name in category_names:

        # Do we have a custom threshold for this category?
        if category_name not in options.category_name_to_threshold:
            print('Warning: category {} in detection file, but not in threshold mapping'.format(
                category_name))
            options.category_name_to_threshold[category_name] = None

        if options.category_name_to_threshold[category_name] is None:
            options.category_name_to_threshold[category_name] = default_threshold

        category_threshold = options.category_name_to_threshold[category_name]
        print('Processing category {} at threshold {}'.format(category_name,category_threshold))

    target_category_names = []
    for c in category_names:

        target_category_names.append(c)

    for combination_length in range(2,len(category_names)+1):

        combined_category_names = list(itertools.combinations(category_names,combination_length))

        for combination in combined_category_names:
            combined_name = '_'.join(combination)
            target_category_names.append(combined_name)

    # Create folder mappings for each category
    for category_name in target_category_names:

        folder_name = category_name

        if category_name in friendly_folder_names:
            folder_name = friendly_folder_names[category_name]

        options.category_name_to_folder[category_name] = \
            os.path.join(options.base_output_folder,folder_name)

    # Create the actual folders
    for folder in options.category_name_to_folder.values():
        os.makedirs(folder,exist_ok=True)

    # Handle species classification thresholds, if specified
    if options.classification_thresholds is not None:

        assert 'classification_categories' in results and \
            results['classification_categories'] is not None, \
            'Classification thresholds specified, but no classification results available'

        classification_categories = results['classification_categories']
        classification_category_name_to_id = {v: k for k, v in classification_categories.items()}
        classification_category_id_to_name = {k: v for k, v in classification_categories.items()}
        options.classification_category_id_to_name = classification_category_id_to_name
        options.classification_categories = classification_categories

        if isinstance(options.classification_thresholds,str):

            # E.g. deer=0.75,cow=0.75
            tokens = options.classification_thresholds.split(',')
            classification_thresholds = {}

            # token = tokens[0]
            for token in tokens:
                subtokens = token.split('=')
                assert (len(subtokens) == 2) and (is_float(subtokens[1])), \
                    'Illegal classification threshold {}'.format(token)
                classification_thresholds[subtokens[0]] = float(subtokens[1])

            # ...for each token

            options.classification_thresholds = classification_thresholds

        # ...if classification thresholds are still in string format

        # Validate the classes in the threshold list
        for class_name in options.classification_thresholds.keys():
            assert class_name in classification_category_name_to_id, \
                'Category {} specified at the command line, but is not available in the results file'.format(
                    class_name)

    # ...if we need to deal with classification categories

    if options.n_threads <= 1 or options.debug_max_images is not None:

        # i_image = 14; im = images[i_image]; im
        for i_image,im in enumerate(tqdm(images)):
            if options.debug_max_images is not None and i_image > options.debug_max_images:
                break
            _process_detections(im,options)
        # ...for each image

    else:

        print('Starting a pool with {} threads'.format(options.n_threads))
        pool = ThreadPool(options.n_threads)
        try:
            process_detections_with_options = partial(_process_detections, options=options)
            _ = list(tqdm(pool.imap(process_detections_with_options, images), total=len(images)))
        finally:
            pool.close()
            pool.join()
            print('Pool closed and joined for folder separation')

    if options.remove_empty_folders:
        print('Removing empty folders from {}'.format(options.base_output_folder))
        remove_empty_folders(options.base_output_folder)

#  ...def separate_detections_into_folders


#%% Interactive driver

if False:

    pass

    #%%

    options = SeparateDetectionsIntoFoldersOptions()

    options.results_file = os.path.expanduser(
        '~/data/snapshot-safari-2022-08-16-KRU-v5a.0.0_detections.json')
    options.base_input_folder = os.path.expanduser('~/data/KRU/KRU_public')
    options.base_output_folder = os.path.expanduser('~/data/KRU-separated')
    options.n_threads = 100
    options.render_boxes = True
    options.allow_existing_directory = True

    #%%

    options = SeparateDetectionsIntoFoldersOptions()

    options.results_file = os.path.expanduser('~/data/ena24-2022-06-15-v5a.0.0_megaclassifier.json')
    options.base_input_folder = os.path.expanduser('~/data/ENA24/images')
    options.base_output_folder = os.path.expanduser('~/data/ENA24-separated')
    options.n_threads = 100
    options.classification_thresholds = 'deer=0.75,cow=0.75,bird=0.75'
    options.render_boxes = True
    options.allow_existing_directory = True

    #%%

    separate_detections_into_folders(options)

    #%% Testing various command-line invocations

    """
    # With boxes, no classification
    python separate_detections_into_folders.py \
        ~/data/ena24-2022-06-15-v5a.0.0_megaclassifier.json \
        ~/data/ENA24/images ~/data/ENA24-separated \
        --threshold 0.17 --animal_threshold 0.2 --n_threads 10 \
        --allow_existing_directory --render_boxes --line_thickness 10 --box_expansion 10

    # No boxes, no classification (default)
    python separate_detections_into_folders.py \
        ~/data/ena24-2022-06-15-v5a.0.0_megaclassifier.json \
        ~/data/ENA24/images ~/data/ENA24-separated \
        --threshold 0.17 --animal_threshold 0.2 --n_threads 10 --allow_existing_directory

    # With boxes, with classification
    python separate_detections_into_folders.py \
        ~/data/ena24-2022-06-15-v5a.0.0_megaclassifier.json ~/data/ENA24/images ~/data/ENA24-separated \
        --threshold 0.17 --animal_threshold 0.2 --n_threads 10 --allow_existing_directory \
        --render_boxes --line_thickness 10 --box_expansion 10 \
        --classification_thresholds "deer=0.75,cow=0.75,bird=0.75"

    # No boxes, with classification
    python separate_detections_into_folders.py \
        ~/data/ena24-2022-06-15-v5a.0.0_megaclassifier.json ~/data/ENA24/images ~/data/ENA24-separated \
        --threshold 0.17 --animal_threshold 0.2 --n_threads 10 --allow_existing_directory \
        --classification_thresholds "deer=0.75,cow=0.75,bird=0.75"
    """

#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser()
    parser.add_argument('results_file', type=str, help='Input .json filename')
    parser.add_argument('base_input_folder', type=str, help='Input image folder')
    parser.add_argument('base_output_folder', type=str, help='Output image folder')

    parser.add_argument('--threshold', type=float, default=None,
                        help='Default confidence threshold for all categories (defaults to ' + \
                            'selection based on model version, other options may override this ' + \
                            'for specific categories)')
    parser.add_argument('--animal_threshold', type=float, default=None,
                        help='Confidence threshold for the animal category')
    parser.add_argument('--human_threshold', type=float, default=None,
                        help='Confidence threshold for the human category')
    parser.add_argument('--vehicle_threshold', type=float, default=None,
                        help='Confidence threshold for vehicle category')
    parser.add_argument('--classification_thresholds', type=str, default=None,
                        help='List of classification thresholds to use for species-based folder ' + \
                             'separation, formatted as, e.g., "deer=0.75,cow=0.75"')

    parser.add_argument('--n_threads', type=int, default=1,
                        help='Number of threads to use for parallel operation (default=1)')

    parser.add_argument('--allow_existing_directory', action='store_true',
                        help='Proceed even if the target directory exists and is not empty')
    parser.add_argument('--no_overwrite', action='store_true',
                        help='Skip images that already exist in the target folder, must also ' + \
                             'specify --allow_existing_directory')
    parser.add_argument('--skip_empty_images', action='store_true',
                        help='Do not copy empty images to the output folder')
    parser.add_argument('--move_images', action='store_true',
                        help='Move images (rather than copying) (not recommended this if you have not ' + \
                             'backed up your data!)')

    parser.add_argument('--render_boxes', action='store_true',
                        help='Render bounding boxes on output images; may result in some ' + \
                             'metadata not being transferred')
    parser.add_argument('--line_thickness', type=int, default=default_line_thickness,
                        help='Line thickness (in pixels) for rendering, only meaningful if ' + \
                             'using render_boxes (defaults to {})'.format(
                             default_line_thickness))
    parser.add_argument('--box_expansion', type=int, default=default_box_expansion,
                        help='Box expansion (in pixels) for rendering, only meaningful if ' + \
                             'using render_boxes (defaults to {})'.format(
                             default_box_expansion))
    parser.add_argument('--category_names_to_blur', type=str, default=None,
                        help='Comma-separated list of category names to blur ' + \
                             '(or a single category name, e.g. "person")')
    parser.add_argument('--remove_empty_folders', action='store_true',
                        help='Remove all empty folders from the target folder at the end of the process, ' + \
                             'whether or not they were created by this script')

    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # Convert to an options object
    options = SeparateDetectionsIntoFoldersOptions()

    args_to_object(args, options)

    def validate_threshold(v,name):
        # print('{} {}'.format(v,name))
        if v is not None:
            assert v >= 0.0 and v <= 1.0, \
                'Illegal {} threshold {}'.format(name,v)

    validate_threshold(args.threshold,'default')
    validate_threshold(args.animal_threshold,'animal')
    validate_threshold(args.vehicle_threshold,'vehicle')
    validate_threshold(args.human_threshold,'human')

    if args.threshold is not None:
        if args.animal_threshold is not None \
            and args.human_threshold is not None \
            and args.vehicle_threshold is not None:
                raise ValueError('Default threshold specified, but all category thresholds ' + \
                                 'also specified... not exactly wrong, but it\'s likely that you ' + \
                                 'meant something else.')

    options.category_name_to_threshold['animal'] = args.animal_threshold
    options.category_name_to_threshold['person'] = args.human_threshold
    options.category_name_to_threshold['vehicle'] = args.vehicle_threshold

    options.overwrite = (not args.no_overwrite)

    separate_detections_into_folders(options)

if __name__ == '__main__':
    main()



#
# Test script for pushing annotations to the eMammal db
#

#%% Imports

import sys
import json
import argparse
import pymysql
import config as cfg

from tqdm import tqdm
from enum import Enum


#%% Database functions

class Categories(Enum):
    animal = 1
    person = 2
    vehicle = 3

mysql_connection = pymysql.connect( host=cfg.host, 
                                user=cfg.username, 
                                passwd=cfg.password, 
                                db=cfg.database, 
                                port=cfg.port)

def update_data(sql):
    with mysql_connection.cursor() as cursor:
        cursor.execute(sql)
        
def get_records_all(sql):
    with mysql_connection.cursor() as cursor:
        sql = sql
        cursor.execute(sql)
        rows = cursor.fetchall()
        return rows

def format_data_print_deployments(rows):
    count = 0
    result = []
    for row in rows:
            count += 1
            print("{}. {}-{}".format(str(count), row[0],row[1]))
            result.append((count, row[0], row[1]))

    return result


#%% Command-line driver

def main(): # noqa
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, help='Input .json filename')
   
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args() 

    print("Enter the number of the deployment:")
    
    rows = get_records_all(''' select * from deployment ''')
    deployments = format_data_print_deployments(rows)
    print("\n")
    deployment_choice = input()
    deployment_id = deployments[int(deployment_choice)][1]
   
    print(deployment_id)

    sql = ''' SELECT emammal_project_taxa_id FROM wild_id.emammal_project_taxa
            where species in ("No Animal", "Unknown Animal", "Homo sapiens", "Vehicle") '''


    emammal_categories = get_records_all(sql)

    with open(args.input_file) as f:
        data = json.load(f)

    images = data['images']
    emammal_category = 0
    for index, im in tqdm(enumerate(images), total=len(images)):      
        fn = im['file']

        if len(im['detections']) <= 0:
            image_type_id = 2
        
            # No-animal category
            emammal_categories = emammal_categories[0]
        else:            
            max_conf = im['max_detection_conf']
            detection = [k for k in im['detections'] if k['conf'] == max_conf]
            category= int(detection[0]['category'])

            if category == Categories.animal:
                image_type_id = 1
                emammal_category = emammal_categories[1]
            else:
                image_type_id = 5
                if category == Categories.person:
                    emammal_category = emammal_categories[2]
                elif category == Categories.vehicle:
                    emammal_category = emammal_categories[3]

            sql = """ UPDATE wild_id.emammal_sequence_annotation,
                      wild_id.image,
                      wild_id.image_sequence,
                      wild_id.deployment
                      SET wild_id.emammal_sequence_annotation.project_taxa_id = 4
                      WHERE wild_id.image.image_sequence_id = wild_id.emammal_sequence_annotation.sequence_id
                      AND wild_id.image_sequence.deployment_id = wild_id.deployment.deployment_id
                      AND wild_id.image.raw_name = '{}' """.format(fn)

            
            print(sql)
            update_data(sql)
            mysql_connection.commit()

if __name__ == '__main__':
    main()




"""

run_detector.py

Module to run an animal detection model on images.  The main function in this script also renders
the predicted bounding boxes on images and saves the resulting images (with bounding boxes).

**This script is not a good way to process lots of images**.  It does not produce a useful
output format, and it does not facilitate checkpointing the results so if it crashes you
would have to start from scratch. **If you want to run a detector on lots of images, you should
check out run_detector_batch.py**.

That said, this script (run_detector.py) is a good way to test our detector on a handful of images
and get super-satisfying, graphical results.

If you would like to *not* use the GPU, set the environment variable CUDA_VISIBLE_DEVICES to "-1".

This script will only consider detections with > 0.005 confidence at all times.
The threshold you provide is only for rendering the results. If you need to
see lower-confidence detections, you can change DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD.

"""

#%% Constants, imports, environment

import argparse
import os
import statistics
import sys
import time
import json
import warnings
import tempfile
import zipfile

import humanfriendly
from tqdm import tqdm

from megadetector.utils import path_utils as path_utils
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.utils.url_utils import download_url
from megadetector.utils.ct_utils import parse_kvp_list
from megadetector.utils.path_utils import compute_file_hash

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)

# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings('ignore', 'Metadata warning', UserWarning)

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# String constants used for consistent reporting of processing errors
FAILURE_INFER = 'inference failure'
FAILURE_IMAGE_OPEN = 'image access failure'

# Number of decimal places to round to for confidence and bbox coordinates
CONF_DIGITS = 3
COORD_DIGITS = 4

# Label mapping for MegaDetector
DEFAULT_DETECTOR_LABEL_MAP = {
    '1': 'animal',
    '2': 'person',
    '3': 'vehicle'  # available in megadetector v4+
}

# Should we allow classes that don't look anything like the MegaDetector classes?
#
# This flag needs to get set if you want to, for example, run an off-the-shelf
# YOLO model with this package.
#
# By default, we error if we see unfamiliar classes.
#
# TODO: the use of a global variable to manage this was fine when this was really
# experimental, but this is really sloppy now that we actually use this code for
# models other than MegaDetector.
USE_MODEL_NATIVE_CLASSES = False

# Detection threshold to recommend to callers when all other mechanisms for choosing
# a model-specific threshold fail
fallback_detection_threshold = 0.2

# Maps a variety of strings that might occur in filenames to canonical version numbers.
#
# Order matters here.
model_string_to_model_version = {

    # Specific model versions that might be expressed in a variety of ways
    'mdv2':'v2.0.0',
    'mdv3':'v3.0.0',
    'mdv4':'v4.1.0',
    'mdv5a':'v5a.0.1',
    'mdv5b':'v5b.0.1',

    'v2':'v2.0.0',
    'v3':'v3.0.0',
    'v4':'v4.1.0',
    'v4.1':'v4.1.0',
    'v5a.0.0':'v5a.0.1',
    'v5b.0.0':'v5b.0.1',
    'v5a.0.1':'v5a.0.1',
    'v5b.0.1':'v5b.0.1',

    'md1000-redwood':'v1000.0.0-redwood',
    'md1000-cedar':'v1000.0.0-cedar',
    'md1000-larch':'v1000.0.0-larch',
    'md1000-sorrel':'v1000.0.0-sorrel',
    'md1000-spruce':'v1000.0.0-spruce',

    'mdv1000-redwood':'v1000.0.0-redwood',
    'mdv1000-cedar':'v1000.0.0-cedar',
    'mdv1000-larch':'v1000.0.0-larch',
    'mdv1000-sorrel':'v1000.0.0-sorrel',
    'mdv1000-spruce':'v1000.0.0-spruce',

    'v1000-redwood':'v1000.0.0-redwood',
    'v1000-cedar':'v1000.0.0-cedar',
    'v1000-larch':'v1000.0.0-larch',
    'v1000-sorrel':'v1000.0.0-sorrel',
    'v1000-spruce':'v1000.0.0-spruce',

    # Arguably less specific model versions
    'redwood':'v1000.0.0-redwood',
    'spruce':'v1000.0.0-spruce',
    'cedar':'v1000.0.0-cedar',
    'larch':'v1000.0.0-larch',

    # Opinionated defaults
    'mdv5':'v5a.0.1',
    'md5':'v5a.0.1',
    'mdv1000':'v1000.0.0-redwood',
    'md1000':'v1000.0.0-redwood',
    'default':'v5a.0.1',
    'megadetector':'v5a.0.1',
}

# python -m http.server 8181
model_url_base = 'https://github.com/agentmorris/MegaDetector/releases/download/v1000.0/'
assert model_url_base.endswith('/')

if os.environ.get('MD_MODEL_URL_BASE') is not None:
    model_url_base = os.environ['MD_MODEL_URL_BASE']
    print('Model URL base provided via environment variable: {}'.format(
        model_url_base
    ))
    if not model_url_base.endswith('/'):
        model_url_base += '/'

# Maps canonical model version numbers to metadata
known_models = {
    'v2.0.0':
    {
        'url':'https://lila.science/public/models/megadetector/megadetector_v2.pb',
        'typical_detection_threshold':0.8,
        'conservative_detection_threshold':0.3,
        'model_type':'tf',
        'normalized_typical_inference_speed':1.0/3.5
    },
    'v3.0.0':
    {
        'url':'https://lila.science/public/models/megadetector/megadetector_v3.pb',
        'typical_detection_threshold':0.8,
        'conservative_detection_threshold':0.3,
        'model_type':'tf',
        'normalized_typical_inference_speed':1.0/3.5
    },
    'v4.1.0':
    {
        'url':'https://github.com/agentmorris/MegaDetector/releases/download/v4.1/md_v4.1.0.pb',
        'typical_detection_threshold':0.8,
        'conservative_detection_threshold':0.3,
        'model_type':'tf',
        'normalized_typical_inference_speed':1.0/3.5
    },
    'v5a.0.0':
    {
        'url':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5a.0.0.pt',
        'typical_detection_threshold':0.2,
        'conservative_detection_threshold':0.05,
        'image_size':1280,
        'model_type':'yolov5',
        'normalized_typical_inference_speed':1.0,
        'md5':'ec1d7603ec8cf642d6e0cd008ba2be8c'
    },
    'v5b.0.0':
    {
        'url':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5b.0.0.pt',
        'typical_detection_threshold':0.2,
        'conservative_detection_threshold':0.05,
        'image_size':1280,
        'model_type':'yolov5',
        'normalized_typical_inference_speed':1.0,
        'md5':'bc235e73f53c5c95e66ea0d1b2cbf542'
    },
    'v5a.0.1':
    {
        'url':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5a.0.1.pt',
        'typical_detection_threshold':0.2,
        'conservative_detection_threshold':0.05,
        'image_size':1280,
        'model_type':'yolov5',
        'normalized_typical_inference_speed':1.0,
        'md5':'60f8e7ec1308554df258ed1f4040bc4f'
    },
    'v5b.0.1':
    {
        'url':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5b.0.1.pt',
        'typical_detection_threshold':0.2,
        'conservative_detection_threshold':0.05,
        'image_size':1280,
        'model_type':'yolov5',
        'normalized_typical_inference_speed':1.0,
        'md5':'f17ed6fedfac2e403606a08c89984905'
    },
    'v1000.0.0-redwood':
    {
        'url':model_url_base + 'md_v1000.0.0-redwood.pt',
        'normalized_typical_inference_speed':1.0,
        'md5':'74474b3aec9cf1a990da38b37ddf9197',
        'typical_detection_threshold':0.3
    },
    'v1000.0.0-spruce':
    {
        'url':model_url_base + 'md_v1000.0.0-spruce.pt',
        'normalized_typical_inference_speed':12.7,
        'md5':'1c9d1d2b3ba54931881471fdd508e6f2'
    },
    'v1000.0.0-larch':
    {
        'url':model_url_base + 'md_v1000.0.0-larch.pt',
        'normalized_typical_inference_speed':2.4,
        'md5':'cab94ebd190c2278e12fb70ffd548b6d'
    },
    'v1000.0.0-cedar':
    {
        'url':model_url_base + 'md_v1000.0.0-cedar.pt',
        'normalized_typical_inference_speed':2.0,
        'md5':'3d6472c9b95ba687b59ebe255f7c576b'
    },
    'v1000.0.0-sorrel':
    {
        'url':model_url_base + 'md_v1000.0.0-sorrel.pt',
        'normalized_typical_inference_speed':7.0,
        'md5':'4339a2c8af7a381f18ded7ac2a4df03e'
    }
}

DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = known_models['v5a.0.0']['typical_detection_threshold']
DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD = 0.005

DEFAULT_BOX_THICKNESS = 4
DEFAULT_BOX_EXPANSION = 0
DEFAULT_LABEL_FONT_SIZE = 16
DETECTION_FILENAME_INSERT = '_detections'

# Approximate inference speeds (in images per second) for MDv5 based on
# benchmarks, only used for reporting very coarse expectations about inference time.
device_token_to_mdv5_inference_speed = {
    '4090':17.6,
    '3090':11.4,
    '3080':9.5,
    '3050':4.2,
    'P2000':2.1,
    # These are written this way because they're MDv4 benchmarks, and MDv5
    # is around 3.5x faster than MDv4.
    'V100':2.79*3.5,
    '2080':2.3*3.5,
    '2060':1.6*3.5
}


#%% Utility functions

def get_detector_metadata_from_version_string(detector_version):
    """
    Given a MegaDetector version string (e.g. "v4.1.0"), returns the metadata for
    the model.  Used for writing standard defaults to batch output files.

    Args:
        detector_version (str): a detection version string, e.g. "v4.1.0", which you
            can extract from a filename using get_detector_version_from_filename()

    Returns:
        dict: metadata for this model, suitable for writing to a MD output file
    """

    if detector_version not in known_models:
        print('Warning: no metadata for unknown detector version {}'.format(detector_version))
        default_detector_metadata = {
            'megadetector_version':'unknown',
            'typical_detection_threshold':0.2,
            'conservative_detection_threshold':0.1
        }
        return default_detector_metadata
    else:
        to_return = known_models[detector_version]
        to_return['megadetector_version'] = detector_version
        return to_return


def get_detector_version_from_filename(detector_filename,
                                       accept_first_match=True,
                                       verbose=False):
    r"""
    Gets the canonical version number string of a detector from the model filename.

    [detector_filename] will almost always end with one of the following:

    * megadetector_v2.pb
    * megadetector_v3.pb
    * megadetector_v4.1 (not produced by run_detector_batch.py, only found in output files from
      the deprecated Azure Batch API)
    * md_v4.1.0.pb
    * md_v5a.0.0.pt
    * md_v5b.0.0.pt

    This function identifies the version number as "v2.0.0", "v3.0.0", "v4.1.0",
    "v4.1.0", "v5a.0.0", and "v5b.0.0", respectively.  See known_models for the list
    of valid version numbers.

    Args:
        detector_filename (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
        accept_first_match (bool, optional): if multiple candidates match the filename, choose the
            first one, otherwise returns the string "multiple"
        verbose (bool, optional): enable additional debug output

    Returns:
        str: a detector version string, e.g. "v5a.0.0", or "multiple" if I'm confused
    """

    fn = os.path.basename(detector_filename).lower()
    matches = []
    for s in model_string_to_model_version.keys():
        if s in fn:
            matches.append(s)
    if len(matches) == 0:
        return 'unknown'
    elif len(matches) > 1:
        if accept_first_match:
            return model_string_to_model_version[matches[0]]
        else:
            if verbose:
                print('Warning: multiple MegaDetector versions for model file {}:'.format(detector_filename))
                for s in matches:
                    print(s)
            return 'multiple'
    else:
        return model_string_to_model_version[matches[0]]


def get_detector_version_from_model_file(detector_filename,verbose=False):
    """
    Gets the canonical detection version from a model file, preferably by reading it
    from the file itself, otherwise based on the filename.

    Args:
        detector_filename (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
        verbose (bool, optional): enable additional debug output

    Returns:
        str: a canonical detector version string, e.g. "v5a.0.0", or "unknown"
    """

    # Try to extract a version string from the filename
    version_string_based_on_filename = get_detector_version_from_filename(
        detector_filename, verbose=verbose)
    if version_string_based_on_filename == 'unknown':
        version_string_based_on_filename = None

    # Try to extract a version string from the file itself; currently this is only
    # a thing for PyTorch models

    version_string_based_on_model_file = None

    if detector_filename.endswith('.pt') or detector_filename.endswith('.zip'):

        from megadetector.detection.pytorch_detector import \
            read_metadata_from_megadetector_model_file
        metadata = read_metadata_from_megadetector_model_file(detector_filename,verbose=verbose)

        if metadata is not None and isinstance(metadata,dict):

            if 'metadata_format_version' not in metadata or \
                not isinstance(metadata['metadata_format_version'],float):

                print(f'Warning: I found a metadata file in detector file {detector_filename}, '+\
                      'but it doesn\'t have a valid format version number')

            elif 'model_version_string' not in metadata or \
                not isinstance(metadata['model_version_string'],str):

                print(f'Warning: I found a metadata file in detector file {detector_filename}, '+\
                      'but it doesn\'t have a format model version string')

            else:

                version_string_based_on_model_file = metadata['model_version_string']

                if version_string_based_on_model_file not in known_models:
                    print('Warning: unknown model version:\n\n{}\n\n...specified in file:\n\n{}'.format(
                        version_string_based_on_model_file,os.path.basename(detector_filename)))

        # ...if there's metadata in this file

    # ...if this looks like a PyTorch file

    # If we got versions strings from the filename *and* the model file...
    if (version_string_based_on_filename is not None) and \
       (version_string_based_on_model_file is not None):

        if version_string_based_on_filename != version_string_based_on_model_file:
            # This is a one-off special case where models were re-released with different filenames
            if (version_string_based_on_filename in ('v5a.0.1','v5b.0.1')) and \
                (version_string_based_on_model_file in ('v5a.0.0','v5b.0.0')):
                pass
            else:
                print(
                    'Warning: model version string in file:' + \
                    '\n\n{}\n\n...is:\n\n{}\n\n...but the filename implies:\n\n{}'.format(
                    os.path.basename(detector_filename),
                    version_string_based_on_model_file,
                    version_string_based_on_filename))

        return version_string_based_on_model_file

    # If we got version string from neither the filename nor the model file...
    if (version_string_based_on_filename is None) and \
       (version_string_based_on_model_file is None):

        print('Warning: could not determine model version string for model file {}'.format(
            detector_filename))
        return None

    elif version_string_based_on_filename is not None:

        return version_string_based_on_filename

    else:

        assert version_string_based_on_model_file is not None
        return version_string_based_on_model_file

# ...def get_detector_version_from_model_file(...)


def estimate_md_images_per_second(model_file, device_name=None):
    r"""
    Estimates how fast MegaDetector will run on a particular device, based on benchmarks.
    Defaults to querying the current device.  Returns None if no data is available for the current
    card/model.  Estimates only available for a small handful of GPUs.  Uses an absurdly simple
    lookup approach, e.g. if the string "4090" appears in the device name, congratulations,
    you have an RTX 4090.

    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
        device_name (str, optional): device name, e.g. blah-blah-4090-blah-blah

    Returns:
        float: the approximate number of images this model version can process on this
        device per second
    """

    if device_name is None:
        try:
            import torch
            device_name = torch.cuda.get_device_name()
        except Exception as e:
            print('Error querying device name: {}'.format(e))
            return None

    # About how fast is this model compared to MDv5?
    model_version = get_detector_version_from_model_file(model_file)

    if model_version not in known_models.keys():
        print('Could not estimate inference speed: error determining model version for model file {}'.format(
            model_file))
        return None

    model_info = known_models[model_version]

    if 'normalized_typical_inference_speed' not in model_info or \
        model_info['normalized_typical_inference_speed'] is None:
        print('No speed ratio available for model type {}'.format(model_version))
        return None

    normalized_inference_speed = model_info['normalized_typical_inference_speed']

    # About how fast would MDv5 run on this device?
    mdv5_inference_speed = None
    for device_token in device_token_to_mdv5_inference_speed.keys():
        if device_token in device_name:
            mdv5_inference_speed = device_token_to_mdv5_inference_speed[device_token]
            break

    if mdv5_inference_speed is None:
        print('No baseline speed estimate available for device {}'.format(device_name))
        return None

    return normalized_inference_speed * mdv5_inference_speed


def get_typical_confidence_threshold_from_results(results):
    """
    Given the .json data loaded from a MD results file, returns a typical confidence
    threshold based on the detector version.

    Args:
        results (dict or str): a dict of MD results, as it would be loaded from a MD results .json
            file, or a .json filename

    Returns:
        float: a sensible default threshold for this model
    """

    # Load results if necessary
    if isinstance(results,str):
        with open(results,'r') as f:
            results = json.load(f)

    default_threshold = None

    # Best case: the .json file tells us the default threshold
    if 'detector_metadata' in results['info'] and \
        'typical_detection_threshold' in results['info']['detector_metadata']:
        default_threshold = results['info']['detector_metadata']['typical_detection_threshold']
    # Worst case: we don't even know what detector this is
    elif ('detector' not in results['info']) or (results['info']['detector'] is None):
        print('Warning: detector version not available in results file, using MDv5 defaults')
        detector_metadata = get_detector_metadata_from_version_string('v5a.0.0')
        default_threshold = detector_metadata['typical_detection_threshold']
    # We know what detector this is, but it doesn't have a default threshold
    # in the .json file
    else:
        print('Warning: detector metadata not available in results file, inferring from MD version')
        try:
            detector_filename = results['info']['detector']
            detector_version = get_detector_version_from_filename(detector_filename)
            detector_metadata = get_detector_metadata_from_version_string(detector_version)
            if 'typical_detection_threshold' in detector_metadata:
                default_threshold = detector_metadata['typical_detection_threshold']
        except Exception:
            pass

    if default_threshold is None:
        print('Could not determine threshold, using fallback threshold of {}'.format(
            fallback_detection_threshold))
        default_threshold = fallback_detection_threshold

    return default_threshold


def is_gpu_available(model_file):
    r"""
    Determines whether a GPU is available, importing PyTorch or TF depending on the extension
    of model_file.  Does not actually load model_file, just uses that to determine how to check
    for GPU availability (PT vs. TF).

    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt

    Returns:
        bool: whether a GPU is available
    """

    if model_file.endswith('.pb'):
        import tensorflow.compat.v1 as tf # type: ignore
        gpu_available = tf.test.is_gpu_available()
        print('TensorFlow version:', tf.__version__)
        print('tf.test.is_gpu_available:', gpu_available)
        return gpu_available
    if not model_file.endswith('.pt'):
        print('Warning: could not determine environment from model file name, assuming PyTorch')

    import torch
    gpu_available = torch.cuda.is_available()
    print('PyTorch reports {} available CUDA devices'.format(torch.cuda.device_count()))
    if not gpu_available:
        try:
            # mps backend only available in torch >= 1.12.0
            if torch.backends.mps.is_built and torch.backends.mps.is_available():
                gpu_available = True
                print('PyTorch reports Metal Performance Shaders are available')
        except AttributeError:
            pass
    return gpu_available


def load_detector(model_file,
                  force_cpu=False,
                  force_model_download=False,
                  detector_options=None,
                  verbose=False):
    r"""
    Loads a TF or PT detector, depending on the extension of model_file.

    Args:
        model_file (str): model filename (e.g. c:/x/z/md_v5a.0.0.pt) or known model
            name (e.g. "MDV5A")
        force_cpu (bool, optional): force the model to run on the CPU even if a GPU
            is available
        force_model_download (bool, optional): force downloading the model file if
            a named model (e.g. "MDV5A") is supplied, even if the local file already
            exists
        detector_options (dict, optional): key/value pairs that are interpreted differently
            by different detectors
        verbose (bool, optional): enable additional debug output

    Returns:
        object: loaded detector object
    """

    # Possibly automatically download the model
    model_file = try_download_known_detector(model_file,
                                             force_download=force_model_download)

    print('GPU available: {}'.format(is_gpu_available(model_file)))

    start_time = time.time()

    if model_file.endswith('.pb'):

        from megadetector.detection.tf_detector import TFDetector
        if force_cpu:
            raise ValueError('force_cpu is not currently supported for TF detectors, ' + \
                             'use CUDA_VISIBLE_DEVICES=-1 instead')
        detector = TFDetector(model_file, detector_options)

    elif model_file.endswith('.pt'):

        from megadetector.detection.pytorch_detector import PTDetector

        # Prepare options specific to the PTDetector class
        if detector_options is None:
            detector_options = {}
        if 'force_cpu' in detector_options:
            if force_cpu != detector_options['force_cpu']:
                print('Warning: over-riding force_cpu parameter ({}) based on detector_options ({})'.format(
                    force_cpu,detector_options['force_cpu']))
        else:
            detector_options['force_cpu'] = force_cpu
        detector_options['use_model_native_classes'] = USE_MODEL_NATIVE_CLASSES
        detector = PTDetector(model_file, detector_options, verbose=verbose)

    else:

        raise ValueError('Unrecognized model format: {}'.format(model_file))

    elapsed = time.time() - start_time

    if verbose:
        print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))

    return detector

# ...def load_detector(...)


#%% Main function

def load_and_run_detector(model_file,
                          image_file_names,
                          output_dir,
                          render_confidence_threshold=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
                          crop_images=False,
                          box_thickness=DEFAULT_BOX_THICKNESS,
                          box_expansion=DEFAULT_BOX_EXPANSION,
                          image_size=None,
                          label_font_size=DEFAULT_LABEL_FONT_SIZE,
                          augment=False,
                          force_model_download=False,
                          detector_options=None,
                          verbose=False):
    r"""
    Loads and runs a detector on target images, and visualizes the results.

    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt, or a known model
            string, e.g. "MDV5A"
        image_file_names (list): list of absolute paths to process
        output_dir (str): folder to write visualized images to
        render_confidence_threshold (float, optional): only render boxes for detections
            above this threshold
        crop_images (bool, optional): whether to crop detected objects to individual images
            (default is to render images with boxes, rather than cropping)
        box_thickness (float, optional): thickness in pixels for box rendering
        box_expansion (float, optional): box expansion in pixels
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing
        label_font_size (float, optional): font size to use for displaying class names
            and confidence values in the rendered images
        augment (bool, optional): enable (implementation-specific) image augmentation
        force_model_download (bool, optional): force downloading the model file if
            a named model (e.g. "MDV5A") is supplied, even if the local file already
            exists
        detector_options (dict, optional): key/value pairs that are interpreted differently
            by different detectors
        verbose (bool, optional): enable additional debug output
    """

    if len(image_file_names) == 0:
        print('Warning: no files available')
        return

    # Possibly automatically download the model
    model_file = try_download_known_detector(model_file,
                                             force_download=force_model_download,
                                             verbose=verbose)

    detector = load_detector(model_file,
                             detector_options=detector_options,
                             verbose=verbose)

    detection_results = []
    time_load = []
    time_infer = []

    # Dictionary mapping output file names to a collision-avoidance count.
    #
    # Since we'll be writing a bunch of files to the same folder, we rename
    # as necessary to avoid collisions.
    output_filename_collision_counts = {}

    def input_file_to_detection_file(fn, crop_index=-1):
        """
        Creates unique file names for output files.

        This function does 3 things:
        1) If the --crop flag is used, then each input image may produce several output
            crops. For example, if foo.jpg has 3 detections, then this function should
            get called 3 times, with crop_index taking on 0, 1, then 2. Each time, this
            function appends crop_index to the filename, resulting in
                foo_crop00_detections.jpg
                foo_crop01_detections.jpg
                foo_crop02_detections.jpg

        2) If the --recursive flag is used, then the same file (base)name may appear
            multiple times. However, we output into a single flat folder. To avoid
            filename collisions, we prepend an integer prefix to duplicate filenames:
                foo_crop00_detections.jpg
                0000_foo_crop00_detections.jpg
                0001_foo_crop00_detections.jpg

        3) Prepends the output directory:
                out_dir/foo_crop00_detections.jpg

        Args:
            fn: str, filename
            crop_index: int, crop number

        Returns: output file path
        """

        fn = os.path.basename(fn).lower()
        name, ext = os.path.splitext(fn)
        if crop_index >= 0:
            name += '_crop{:0>2d}'.format(crop_index)
        fn = '{}{}{}'.format(name, DETECTION_FILENAME_INSERT, '.jpg')
        if fn in output_filename_collision_counts:
            n_collisions = output_filename_collision_counts[fn]
            fn_original = fn
            fn = '{:0>4d}'.format(n_collisions) + '_' + fn
            output_filename_collision_counts[fn_original] += 1
        else:
            output_filename_collision_counts[fn] = 0
        fn = os.path.join(output_dir, fn)
        return fn

    # ...def input_file_to_detection_file()

    for im_file in tqdm(image_file_names):

        try:
            start_time = time.time()

            image = vis_utils.load_image(im_file)

            elapsed = time.time() - start_time
            time_load.append(elapsed)

        except Exception as e:
            print('Image {} cannot be loaded, error: {}'.format(im_file, str(e)))
            result = {
                'file': im_file,
                'failure': FAILURE_IMAGE_OPEN
            }
            detection_results.append(result)
            continue

        try:
            start_time = time.time()

            result = detector.generate_detections_one_image(
                image,
                im_file,
                detection_threshold=DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                image_size=image_size,
                augment=augment)
            detection_results.append(result)

            elapsed = time.time() - start_time
            time_infer.append(elapsed)

        except Exception as e:
            print('An error occurred while running the detector on image {}: {}'.format(
                im_file, str(e)))
            continue

        try:
            if crop_images:

                images_cropped = vis_utils.crop_image(result['detections'], image,
                                   confidence_threshold=render_confidence_threshold,
                                   expansion=box_expansion)

                for i_crop, cropped_image in enumerate(images_cropped):
                    output_full_path = input_file_to_detection_file(im_file, i_crop)
                    cropped_image.save(output_full_path)

            else:

                # Image is modified in place
                vis_utils.render_detection_bounding_boxes(result['detections'], image,
                            label_map=DEFAULT_DETECTOR_LABEL_MAP,
                            confidence_threshold=render_confidence_threshold,
                            thickness=box_thickness, expansion=box_expansion,
                            label_font_size=label_font_size,
                            box_sort_order='confidence')
                output_full_path = input_file_to_detection_file(im_file)
                image.save(output_full_path)

        except Exception as e:
            print('Visualizing results on the image {} failed. Exception: {}'.format(im_file, e))
            continue

    # ...for each image

    ave_time_load = statistics.mean(time_load)
    ave_time_infer = statistics.mean(time_infer)
    if len(time_load) > 1 and len(time_infer) > 1:
        std_dev_time_load = humanfriendly.format_timespan(statistics.stdev(time_load))
        std_dev_time_infer = humanfriendly.format_timespan(statistics.stdev(time_infer))
    else:
        std_dev_time_load = 'not available (<=1 image processed)'
        std_dev_time_infer = 'not available (<=1 image processed)'
    print('On average, for each image,')
    print('- loading took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_load),
                                                    std_dev_time_load))
    print('- inference took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_infer),
                                                      std_dev_time_infer))

# ...def load_and_run_detector()


def _validate_zip_file(file_path, file_description='file'):
    """
    Validates that a .pt file is a valid zip file.

    Args:
        file_path (str): path to the file to validate
        file_description (str): descriptive string for error messages

    Returns:
        bool: True if valid, False otherwise
    """
    try:
        with zipfile.ZipFile(file_path, 'r') as zipf:
            corrupt_file = zipf.testzip()
            if corrupt_file is not None:
                print('{} {} contains at least one corrupt file: {}'.format(
                    file_description.capitalize(), file_path, corrupt_file))
                return False
        return True
    except (zipfile.BadZipFile, zipfile.LargeZipFile) as e:
        print('{} {} appears to be corrupted (bad zip): {}'.format(
            file_description.capitalize(), file_path, str(e)))
        return False
    except Exception as e:
        print('Error validating {}: {}'.format(file_description, str(e)))
        return False


def _validate_md5_hash(file_path, expected_hash, file_description='file'):
    """
    Validates that a file has the expected MD5 hash.

    Args:
        file_path (str): path to the file to validate
        expected_hash (str): expected MD5 hash
        file_description (str): descriptive string for error messages

    Returns:
        bool: True if hash matches, False otherwise
    """
    try:
        actual_hash = compute_file_hash(file_path, algorithm='md5').lower()
        expected_hash = expected_hash.lower()
        if actual_hash != expected_hash:
            print('{} {} has incorrect hash. Expected: {}, Actual: {}'.format(
                file_description.capitalize(), file_path, expected_hash, actual_hash))
            return False
        return True
    except Exception as e:
        print('Error computing hash for {}: {}'.format(file_description, str(e)))
        return False


def _download_model(model_name,force_download=False):
    """
    Downloads one of the known models to local temp space if it hasn't already been downloaded.

    Args:
        model_name (str): a known model string, e.g. "MDV5A".  Returns None if this string is not
            a known model name.
        force_download (bool, optional): whether to download the model even if the local target
            file already exists
    """

    model_tempdir = os.path.join(tempfile.gettempdir(), 'megadetector_models')
    os.makedirs(model_tempdir,exist_ok=True)

    # This is a lazy fix to an issue... if multiple users run this script, the
    # "megadetector_models" folder is owned by the first person who creates it, and others
    # can't write to it.  I could create uniquely-named folders, but I philosophically prefer
    # to put all the individual UUID-named folders within a larger folder, so as to be a
    # good tempdir citizen.  So, the lazy fix is to make this world-writable.
    try:
        os.chmod(model_tempdir,0o777)
    except Exception:
        pass
    if model_name.lower() not in known_models:
        print('Unrecognized downloadable model {}'.format(model_name))
        return None

    model_info = known_models[model_name.lower()]
    url = model_info['url']
    destination_filename = os.path.join(model_tempdir,url.split('/')[-1])

    # Check whether the file already exists, in which case we want to validate it
    if os.path.exists(destination_filename) and not force_download:

        # Only validate .pt files, not .pb files
        if destination_filename.endswith('.pt'):

            is_valid = True

            # Check whether the file is a valid zip file (.pt files are zip files in disguise)
            if not _validate_zip_file(destination_filename,
                                      'existing model file'):
                is_valid = False

            # Check MD5 hash if available
            if is_valid and \
                ('md5' in model_info) and \
                (model_info['md5'] is not None) and \
                (len(model_info['md5'].strip()) > 0):

                if not _validate_md5_hash(destination_filename, model_info['md5'],
                                          'existing model file'):
                    is_valid = False

            # If validation failed, delete the corrupted file and re-download
            if not is_valid:
                print('Deleting corrupted model file and re-downloading: {}'.format(
                    destination_filename))
                try:
                    os.remove(destination_filename)
                    # This should be a no-op at this point, but it can't hurt
                    force_download = True
                except Exception as e:
                    print('Warning: failed to delete corrupted file {}: {}'.format(
                        destination_filename, str(e)))
                    # Continue with download attempt anyway, setting force_download to True
                    force_download = True
            else:
                print('Model {} already exists and is valid at {}'.format(
                    model_name, destination_filename))
                return destination_filename

    # Download the model
    try:
        local_file = download_url(url,
                                destination_filename=destination_filename,
                                progress_updater=None,
                                force_download=force_download,
                                verbose=True)
    except Exception as e:
        print('Error downloading model {} from {}: {}'.format(model_name, url, str(e)))
        raise

    # Validate the downloaded file if it's a .pt file
    if local_file and local_file.endswith('.pt'):

        # Check if the downloaded file is a valid zip file
        if not _validate_zip_file(local_file, "downloaded model file"):
            # Clean up the corrupted download
            try:
                os.remove(local_file)
            except Exception:
                pass
            return None

        # Check MD5 hash if available
        if ('md5' in model_info) and \
            (model_info['md5'] is not None) and \
            (len(model_info['md5'].strip()) > 0):

            if not _validate_md5_hash(local_file, model_info['md5'], "downloaded model file"):
                # Clean up the corrupted download
                try:
                    os.remove(local_file)
                except Exception:
                    pass
                return None

    print('Model {} available at {}'.format(model_name,local_file))
    return local_file

# ...def _download_model(...)

def try_download_known_detector(detector_file,force_download=False,verbose=False):
    """
    Checks whether detector_file is really the name of a known model, in which case we will
    either read the actual filename from the corresponding environment variable or download
    (if necessary) to local temp space.  Otherwise just returns the input string.

    Args:
        detector_file (str): a known model string (e.g. "MDV5A"), or any other string (in which
            case this function is a no-op)
        force_download (bool, optional): whether to download the model even if the local target
            file already exists
        verbose (bool, optional): enable additional debug output

    Returns:
        str: the local filename to which the model was downloaded, or the same string that
        was passed in, if it's not recognized as a well-known model name
    """

    model_string = detector_file.lower()

    # If this is a short model string (e.g. "MDV5A"), convert to a canonical version
    # string (e.g. "v5a.0.0")
    if model_string in model_string_to_model_version:

        if verbose:
            print('Converting short string {} to canonical version string {}'.format(
                model_string,
                model_string_to_model_version[model_string]))
        model_string = model_string_to_model_version[model_string]

    if model_string in known_models:

        if detector_file in os.environ:
            fn = os.environ[detector_file]
            print('Reading MD location from environment variable {}: {}'.format(
                detector_file,fn))
            detector_file = fn
        else:
            detector_file = _download_model(model_string,force_download=force_download)

    return detector_file




#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser(
        description='Module to run an animal detection model on images')

    parser.add_argument(
        'detector_file',
        help='Path detector model file (.pb or .pt).  Can also be MDV4, MDV5A, or MDV5B to request automatic download.')

    # Must specify either an image file or a directory
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--image_file',
        type=str,
        default=None,
        help='Single file to process, mutually exclusive with --image_dir')
    group.add_argument(
        '--image_dir',
        type=str,
        default=None,
        help='Directory to search for images, with optional recursion by adding --recursive')

    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Recurse into directories, only meaningful if using --image_dir')

    parser.add_argument(
        '--output_dir',
        type=str,
        default=None,
        help='Directory for output images (defaults to same as input)')

    parser.add_argument(
        '--image_size',
        type=int,
        default=None,
        help=('Force image resizing to a (square) integer size (not recommended to change this)'))

    parser.add_argument(
        '--threshold',
        type=float,
        default=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
        help=('Confidence threshold between 0 and 1.0; only render' +
              ' boxes above this confidence (defaults to {})'.format(
              DEFAULT_RENDERING_CONFIDENCE_THRESHOLD)))

    parser.add_argument(
        '--crop',
        default=False,
        action='store_true',
        help=('If set, produces separate output images for each crop, '
              'rather than adding bounding boxes to the original image'))

    parser.add_argument(
        '--augment',
        default=False,
        action='store_true',
        help=('Enable image augmentation'))

    parser.add_argument(
        '--box_thickness',
        type=int,
        default=DEFAULT_BOX_THICKNESS,
        help=('Line width (in pixels) for box rendering (defaults to {})'.format(
              DEFAULT_BOX_THICKNESS)))

    parser.add_argument(
        '--box_expansion',
        type=int,
        default=DEFAULT_BOX_EXPANSION,
        help=('Number of pixels to expand boxes by (defaults to {})'.format(
              DEFAULT_BOX_EXPANSION)))

    parser.add_argument(
        '--label_font_size',
        type=int,
        default=DEFAULT_LABEL_FONT_SIZE,
        help=('Label font size (defaults to {})'.format(
              DEFAULT_LABEL_FONT_SIZE)))

    parser.add_argument(
        '--process_likely_output_images',
        action='store_true',
        help=('By default, we skip images that end in {}, because they probably came from this script. '\
              .format(DETECTION_FILENAME_INSERT) + \
              'This option disables that behavior.'))

    parser.add_argument(
        '--force_model_download',
        action='store_true',
        help=('If a named model (e.g. "MDV5A") is supplied, force a download of that model even if the ' +\
              'local file already exists.'))

    parser.add_argument(
        '--verbose',
        action='store_true',
        help=('Enable additional debug output'))

    parser.add_argument(
        '--detector_options',
        nargs='*',
        metavar='KEY=VALUE',
        default='',
        help='Detector-specific options, as a space-separated list of key-value pairs')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    detector_options = parse_kvp_list(args.detector_options)

    # If the specified detector file is really the name of a known model, find
    # (and possibly download) that model
    args.detector_file = try_download_known_detector(args.detector_file,
                                                     force_download=args.force_model_download)

    assert os.path.exists(args.detector_file), 'detector file {} does not exist'.format(
        args.detector_file)
    assert 0.0 < args.threshold <= 1.0, 'Confidence threshold needs to be between 0 and 1'

    if args.image_file:
        image_file_names = [args.image_file]
    else:
        image_file_names = path_utils.find_images(args.image_dir, args.recursive)

    # Optionally skip images that were probably generated by this script
    if not args.process_likely_output_images:
        image_file_names_valid = []
        for fn in image_file_names:
            if os.path.splitext(fn)[0].endswith(DETECTION_FILENAME_INSERT):
                print('Skipping likely output image {}'.format(fn))
            else:
                image_file_names_valid.append(fn)
        image_file_names = image_file_names_valid

    print('Running detector on {} images...'.format(len(image_file_names)))

    if args.output_dir:
        os.makedirs(args.output_dir, exist_ok=True)
    else:
        if args.image_dir:
            args.output_dir = args.image_dir
        else:
            # but for a single image, args.image_dir is also None
            args.output_dir = os.path.dirname(args.image_file)

    load_and_run_detector(model_file=args.detector_file,
                          image_file_names=image_file_names,
                          output_dir=args.output_dir,
                          render_confidence_threshold=args.threshold,
                          box_thickness=args.box_thickness,
                          box_expansion=args.box_expansion,
                          crop_images=args.crop,
                          image_size=args.image_size,
                          label_font_size=args.label_font_size,
                          augment=args.augment,
                          # If --force_model_download was specified, we already handled it
                          force_model_download=False,
                          detector_options=detector_options,
                          verbose=args.verbose)

if __name__ == '__main__':
    main()


#%% Interactive driver(s)

if False:

    pass

    #%% Test model download

    r"""
    cd i:\models\all_models_in_the_wild
    i:
    python -m http.server 8181
    """

    model_name = 'redwood'
    try_download_known_detector(model_name,force_download=True,verbose=True)


    #%% Load and run detector

    model_file = r'c:\temp\models\md_v4.1.0.pb'
    image_file_names = path_utils.find_images(r'c:\temp\demo_images\ssverymini')
    output_dir = r'c:\temp\demo_images\ssverymini'
    render_confidence_threshold = 0.8
    crop_images = True

    load_and_run_detector(model_file=model_file,
                          image_file_names=image_file_names,
                          output_dir=output_dir,
                          render_confidence_threshold=render_confidence_threshold,
                          crop_images=crop_images)
